package com.blackledge.david.html;

import java.awt.Color;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Deque;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.text.MutableAttributeSet;
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTML.Tag;
import javax.swing.text.html.HTMLEditorKit.ParserCallback;

// JSOUP
import org.jsoup.nodes.Attribute;
import org.jsoup.nodes.Attributes;
import org.jsoup.nodes.Element;
import org.jsoup.nodes.Node;
import org.jsoup.nodes.TextNode;
import org.jsoup.select.NodeVisitor;
// /JSOUP

import com.blackledge.david.html.hme.VimeoHandler;
import com.blackledge.david.html.hme.YouTubeLauncher;

public class HTMLParserCallback extends ParserCallback 
// JSOUP
implements NodeVisitor 
// /JSOUP
{
		public class Styles {

			private Color color;
			private Color bgcolor;
			private String bg;
			private Boolean bold;
			private Boolean italic;
			private Integer size;
			private String font;
			private Styles parent;
			private String list;
			private boolean isBlock;
			private Integer align;
			private Integer bgposition;
			private Integer bgrepeat;

			public Styles(Color color, Boolean bold, Boolean italic,
					Integer size, String font, String list, Styles parent, Color bgcolor, String bg, Integer bgrepeat, Integer bgposition, Integer align, boolean isBlock) {
						this.color = color;
						this.bold = bold;
						this.italic = italic;
						this.size = size;
						this.font = font;
						this.list = list;
						this.parent = parent;
						this.bgcolor = bgcolor;
						this.bg = bg;
						this.bgrepeat = bgrepeat;
						this.bgposition = bgposition;
						this.align = align;
						this.isBlock = isBlock;
			}
			
			public Boolean getBold() {
				if(bold == null && parent != null)
					return parent.getBold();
//				if(bold!=null && bold)debugOut("BOLD");
				return bold;
			}
			public Color getColor() {
				if(color == null && parent != null)
					return parent.getColor();
//				if(color != null)debugOut("COLOR "+color);
				return color;
			}
			public Color getBlockBGColor() {
				Styles s = this;
				while(s != null && !s.isBlock ) {
					s = s.parent;
				}
				if(s == null) return null;
				return s.getBGColor();
			}
			public String getBlockBG() {
				Styles s = this;
				while(s != null && !s.isBlock ) {
					s = s.parent;
				}
				if(s == null) return null;
				return s.getBG();
			}
			public Color getBGColor() {
				// only look to parent if we're a block, or we're not going up to a block from a non-block
				if(bgcolor == null && parent != null && (this.isBlock || !parent.isBlock) )
					return parent.getBGColor();
				return bgcolor;
			}
			public String getBG() {
				// only look to parent if we're a block, or we're not going up to a block from a non-block
				if(bg == null && parent != null && (this.isBlock || !parent.isBlock) )
					return parent.getBG();
				return bg;
			}
			public Integer getBlockAlign() {
				Styles s = this;
				while(s != null && !s.isBlock ) {
					s = s.parent;
				}
				if(s == null) return null;
				// only look to parent if we're a block, or we're not going up to a block from a non-block
				if(align == null && parent != null && (this.isBlock || !parent.isBlock) )
					return parent.getBlockAlign();
				return align;
			}
			public String getFont() {
				if(font == null && parent != null)
					return parent.getFont();
//				if(font != null)debugOut("FONT "+font);
				return font;
			}
			public Boolean getItalic() {
				if(italic == null && parent != null)
					return parent.getItalic();
//				if(italic!=null && italic)debugOut("ITALIC");
				return italic;
			}
			public Integer getSize() {
				if(size == null && parent != null)
					return parent.getSize();
//				if(size != null)debugOut("SIZE "+size);
				return size;
			}
			public String getListStyle() {
				// have to merge image and non-image styles
				
				// if current style is none or a url, that's the most important thing and we can return.
				if(list != null) {
					if(list.toLowerCase().contains("://") || list.trim().equalsIgnoreCase("none")) {
						return list;
					}
				}
				String parentStyle = null;
				if(parent != null)
					parentStyle = parent.getListStyle();
				if(list != null && parentStyle != null) {
					// if parent is "none" then there is nothing to merge with current style
					if(parentStyle.trim().equalsIgnoreCase("none")) {
						return list;
					}
					if((parentStyle.toLowerCase().contains("://"))) {
						// FIXME should really merge, not replace, but caller doesn't support it yet
						return parentStyle ;//+ " "+list;
					}
				}
				if(list == null) return parentStyle;
//				if(list != null)debugOut("getListStyle "+list);
				return list;
			}
			
			@Override
			public String toString() {
				return "Styles:"+font+"-"+size+"-bold"+bold+"-italic"+italic+" "+color+" "+list;
			}

			public Integer getBlockBGRepeat() {
				Styles s = this;
				while(s != null && !s.isBlock ) {
					s = s.parent;
				}
				if(s == null) return null;
				// only look to parent if we're a block, or we're not going up to a block from a non-block
				if(bgrepeat == null && parent != null && (this.isBlock || !parent.isBlock) )
					return parent.getBlockBGRepeat();
				return bgrepeat;
			}
			
			public Integer getBlockBGPosition() {
				Styles s = this;
				while(s != null && !s.isBlock ) {
					s = s.parent;
				}
				if(s == null) return null;
				// only look to parent if we're a block, or we're not going up to a block from a non-block
				if(bgposition == null && parent != null && (this.isBlock || !parent.isBlock) )
					return parent.getBlockBGPosition();
				return bgposition;
			}

			public Integer getBGRepeat() {
				// only look to parent if we're a block, or we're not going up to a block from a non-block
				if(bgrepeat == null && parent != null && (this.isBlock || !parent.isBlock) )
					return parent.getBGRepeat();
				return bgrepeat;
			}

			public Integer getBGPosition() {
				// only look to parent if we're a block, or we're not going up to a block from a non-block
				if(bgposition == null && parent != null && (this.isBlock || !parent.isBlock) )
					return parent.getBGPosition();
				return bgposition;
			}
			

	}

		private static final String TITLE_FLAG_VALUE = "GIVE ME A TITLE";
		
		public Color bodyColor = null;
		public Color fgColor = null;
		public String bodyImg = null;
		public int bodyBgposition = HTMLRow.ROW_BG_POSITION_DEFAULT;
		public int bodyBgrepeat = HTMLRow.ROW_BG_REPEAT_DEFAULT;
		public Color linkColor = null;

		private static int DEFAULT_TEXT_SIZE=15;
		
		private HTMLRow currentRow;
		public List<HTMLRow> rows;
		boolean inBody = false;
		private String ignoreUntil;
		private URL urlBase;
		private String linkSrc;
		private int ignoreDepth;
		
		public String title;
		public String refresh;
		public String embeddedStream;
		//TODO is there some way to share definitions between this and YouTubeLauncher and VimeoLauncher?
		private Pattern embeddedVideoPattern = Pattern.compile(
				"(" +
				".*youtu(be\\.com/(watch|v/|embed/)|\\.be/).*" +
				"|" +
				".*vimeo\\.com/(?:video/)?(\\d+).*" +
				"|" +
				".*vine.co/v/.*/embed.*" +
				")");
		/** the row index of the first "mainContentOfPage" entry, if any. */
		public int mainContentRow = -1;
		public String[] accessNumLinks = new String[10];
		
		private int currentlyCombiningLinks = 0;
		private HTMLStyleChunk combiningLinkSample;
		private static int maxCombinedLinks = 5;

		private static Pattern splitStyleValuesPattern;

		private static Integer debuglevel;
		private boolean groupIndividualLinks;
		private String linkSeparator;
		
		private String bulletImage;
		private String circleImage;
		private String squareImage;
		private boolean inList = false;
		
		private int formIndex = 0; // unique-ize the forms		
		private String form;
		private String form_textarea = null;
		private String select;
		private String option_value;
		private String buttonName;
		
		public URL icon;
		private boolean preFormatted;
		
		private int tableIndex;
		private LinkedList<String> currentTable = new LinkedList<String>();
		private boolean isLayingOutTables = true;

		private Object select_attributes;
		private Object textarea_attributes;
		
		public HTMLParserCallback(URL urlBase, boolean groupIndividualLinks, String linkSeparator, String bulletImage, String circleImage, String squareImage, boolean layOutTables) {
			super();
			this.urlBase = urlBase;
			this.groupIndividualLinks = groupIndividualLinks;
			this.linkSeparator = linkSeparator;
			this.bulletImage = bulletImage;
			this.circleImage = circleImage;
			this.squareImage = squareImage;
			
			this.isLayingOutTables = layOutTables;
			
			this.rows = Collections.synchronizedList( new ArrayList<HTMLRow>() );
			newRow();
		}
		
		@Override
		public void handleStartTag(Tag t, MutableAttributeSet a, int pos) {
			
//			super.handleStartTag(t, a, pos);
			if(t == null) return;
			//debugOut(t);
			String tagName = t.toString().toUpperCase();
			handleStartTag(tagName, t.breaksFlow(), a);
		}
		
		private void handleStartTag(String tagName, boolean breaksFlow,
				Object a) {
			if (ignoreUntil != null) {
				if (tagName.equals(ignoreUntil)) {
					ignoreDepth++;
					return;
				}
			}
			// handle style="display:none" (with nested tags) for ignore.
			// any tag, style="display:none"

			// Track whether a bullet is needed at beginning of line
			boolean addBullet = false;
//			if(tagName.equals("DIV")) {
//				debugOut("");
//			}
			Object s = getAttributeValue(a, "style");
//			debugOut(s);
			String displayValue = null;
			if(s != null) {
				String style = String.valueOf(s);
				if (style.trim().length() == 0) {
					pushStyle(null, breaksFlow);
				} else {
					displayValue = getStyleValue(style, "display");
					if(displayValue != null && displayValue.equalsIgnoreCase("none")) {
	//				if(style.matches(".*display\\:\\s*none.*")) {
	//debugOut(style);
						if(ignoreUntil == null) {
							ignoreUntil = tagName;
							ignoreDepth = 1;
							return;
						}
					}
					breaksFlow = isDisplayBlock(displayValue, breaksFlow);
					pushStyle(style, breaksFlow);
				}
			} else {
				pushStyle(null, breaksFlow);
			}
			
			// handle "display"-based work.
			if ((displayValue != null && displayValue.equalsIgnoreCase("list-item"))
					|| (displayValue == null && tagName.equals("LI"))) {
//				debugOut(tagName+" display:"+displayValue);
				addBullet = true;
			} else {
				addBullet = false;
			}
			//TODO add here and at endTag handling of block vs. inline regardless of tagName
			
			if (mainContentRow < 0) {
				Object rel = getAttributeValue(a, "rel");
				if (rel != null) {
					if ("mainContentOfPage".equals(String.valueOf(rel)))
						mainContentRow = getRows().size() - 1;
				}

				Object itemprop = getAttributeValue(a, "itemprop");
				if (itemprop != null) {
					if ("mainContentOfPage".equals(String.valueOf(itemprop)))
						mainContentRow = getRows().size() - 1;
				}
				if(mainContentRow >=0) {
					debugOut("found mainContentOfPage at row "+mainContentRow);
				}
			}

			//ID's on anything count as anchors for urls
			Object id = getAttributeValue(a, "id");
			if(id != null && !id.equals("")) {
				setAnchor(String.valueOf(id));
//			} else { // why was I doing this? default anchor should be OK, but null could clear a named anchor
//				setAnchor(null);
			}

			if(tagName.equals("BODY")) {
				inBody = true;
				// background="images/CharleneDrawing1.jpg" bgcolor=#FFA0A0 
				// link="purple" alink="yellow" vlink="gray" 
//FIXME				HTML.Attribute.BGCOLOR
				Enumeration<?> atts = getAttributeEnumeration(a);
				while(atts.hasMoreElements()){
					Object key = atts.nextElement();
					if(key.toString().equalsIgnoreCase("bgcolor")) {
						Object val = getAttributeValue(a, key);
						bodyColor = getColor( String.valueOf(val));
					}
					else if(key.toString().equalsIgnoreCase("color")) {
						Object val = getAttributeValue(a, key);
						fgColor = getColor( String.valueOf(val));
					}
					else if(key.toString().equalsIgnoreCase("link")) {
						Object val = getAttributeValue(a, key);
						linkColor = getColor( String.valueOf(val));
					}
					else if(key.toString().equalsIgnoreCase("background")) {
						Object val = getAttributeValue(a, key);
						bodyImg =  String.valueOf(val);
					}
					else if(key.toString().equalsIgnoreCase("style")) {
						String style = getAttributeString(a, key.toString());
						if(fgColor == null) {
							Color color1 = getStyleColor(style);
							if(color1 != null) {
								fgColor = color1;
							}
						}
						
						if(bodyImg == null) {
							String bodyImg1 = getStyleBGImage(style);
							if(bodyImg1 != null)
								bodyImg = bodyImg1;
						}
						Integer pos = getStyleBGPosition(style);
						if(pos != null) bodyBgposition = pos;
						Integer rep = getStyleBGRepeat(style);
						if(rep != null) bodyBgrepeat = rep;
						
						// could also set bgcolor I think
						if(bodyColor == null) {
							Color bodyColor1 = getStyleBGColor(style);
							if(bodyColor1 != null)
								bodyColor = bodyColor1;
						}
					}
					if(bodyColor == null) {
						bodyColor = Color.white;
					}
					if(fgColor == null) {
						fgColor = Color.black;
					}
					if(linkColor == null) {
						linkColor = Color.blue;
					}
					
					// add body's style to stack - do the same for other things that have alternate attributes
					addStyle(null, null, fgColor, null, null, null, bodyColor, bodyImg, null);
					
				}


			}
			if(tagName.equals("TITLE")) {
				this.title = TITLE_FLAG_VALUE;
			}
			if(breaksFlow && inBody) {
				newRow();
			}
			
			if(addBullet) {
				addListBullet();
			}
//			if(tagName.equals("H2")) {
//				debugOut(tagName);
//			}
			if(tagName.equals("FRAMESET")) {
				// often get a frameset without a body
				inBody = true;
				newRow();
			} else if(tagName.equals("NOFRAMES")) {
				// often get a frameset without a body
				inBody = true;
			} else if(tagName.equals("FRAME") || tagName.equals("IFRAME")) {
				handleFrameTag(tagName, a);
				if(tagName.equals("IFRAME")) {
					ignoreUntil = tagName;
				}
			} else if(tagName.equals("VIDEO")) {
				handleVideoTag(a);
			} else if(tagName.equals("AUDIO")) {
				handleAudioTag(a);
			} else if(tagName.equals("SOURCE")) {
				// optional src alternatives for VIDEO and AUDIO tag
				handleSourceTag(a);
			}
			if(tagName.equals("UL")) {
				inList = true;
			} else if(tagName.equals("LI")) {
			// We do this earlier, independent of the element because anything
			// can be list-item display type and LI can be not list-item
//				addListBullet();
				//TODO this is where we would addStyle based on type="A" or type="disc"
			} else if(tagName.equals("SELECT")) {
				String selected = getAttributeString(a, "defaultSelected");
				select_attributes = a;
				select = "" + selected;
			} else if(tagName.equals("OPTION")) {
				String value=null;
				boolean selected=false;
				Enumeration<?> atts = getAttributeEnumeration(a);
				while (atts.hasMoreElements()) {
					Object key = atts.nextElement();
					if (key.toString().equalsIgnoreCase("value")) {
						Object val = getAttributeValue(a, key);
						if (val != null)
							value = String.valueOf(val);
					} else if (key.toString().equalsIgnoreCase("selected")) {
						selected = true;
					}
				}
				// option_value=null is a flag, we need to negate that flag when an
				// option has no value at all
				if(value == null) {
					// but create a unique value so our code will "hold" a selection
					value = "null - "+Math.random();
				}
				option_value = value;
				if(selected) {
					setSelected(value);
				}
			} else if(tagName.equals("FORM")) {
				++formIndex;
				String submitType = null;
				String url = null;
				Enumeration<?> atts = getAttributeEnumeration(a);
				while (atts.hasMoreElements()) {
					Object key = atts.nextElement();
					if (key.toString().equalsIgnoreCase("action")) {
						Object val = getAttributeValue(a, key);
						if (val != null)
							url = String.valueOf(val);
					} else if (key.toString().equalsIgnoreCase("method")) {
						Object val = getAttributeValue(a, key);
						if (val != null)
							submitType = String.valueOf(val);
					}
				}
				if(url == null) url = "";
				form = formIndex+"&"+submitType+":"+getSrc(urlBase, url);
			} else if(tagName.equals("TEXTAREA")) {
				String name = ""; // default to non-null - it's really a textarea
				Enumeration<?> atts = getAttributeEnumeration(a);
				String cols = "20";
				String rows = "2";
				while (atts.hasMoreElements()) {
					Object key = atts.nextElement();
					if (key.toString().equalsIgnoreCase("name")) {
						Object val = getAttributeValue(a, key);
						if (val != null)
							name = String.valueOf(val);
					} else if (key.toString().equalsIgnoreCase("cols")) {
						Object val = getAttributeValue(a, key);
						if (val != null)
							cols = String.valueOf(val);
					} else if (key.toString().equalsIgnoreCase("rows")) {
						Object val = getAttributeValue(a, key);
						if (val != null)
							rows = String.valueOf(val);
					}

				}
				textarea_attributes = a;
				form_textarea = name+"|"+cols+"x"+rows;
//				preFormatted = true; no idea why this was here
			} else if(tagName.equals("FONT")) {
				Enumeration<?> atts = getAttributeEnumeration(a);
				Color fontTagColor = null;
				String fontTagFont = null;
				Integer fontTagSize = null;
				while(atts.hasMoreElements()){
					Object key = atts.nextElement();
					if(key.toString().equalsIgnoreCase("color")) {
						Object val = getAttributeValue(a, key);
//						debugOut("color val:"+val);
//						if(key instanceof AttributeSet.ColorAttribute) {
//							try {
//							fontTagColor = (Color) val;
//							} catch (ClassCastException e) {
//								debugOut("color att is "+val.getClass());
//							}
//						} else {
							fontTagColor = getColor( String.valueOf(val));
//						}
					} else if(key.toString().equalsIgnoreCase("size")) {
						try {
							fontTagSize = Integer.parseInt(String.valueOf(getAttributeValue(a, key)));
							// actually FONT size is a value from 1-7, where 3 is "normal"
							switch(fontTagSize) {
							case 1:
								fontTagSize = 10;
								break;
							case 2:
								fontTagSize = 13;
								break;
							case 3:
								fontTagSize = DEFAULT_TEXT_SIZE;
								break;
							case 4:
								fontTagSize = 17;
								break;
							case 5:
								fontTagSize = 22;
								break;
							case 6:
								fontTagSize = 26;
								break;
							case 7:
								fontTagSize = 30;
								break;
							default:
								fontTagSize = DEFAULT_TEXT_SIZE;
								break;
							}
						} catch (Exception e) {
							fontTagSize = DEFAULT_TEXT_SIZE;
						}
					} else if(key.toString().equalsIgnoreCase("font")) {
						fontTagFont = (String) getAttributeValue(a, key);
					}
//					debugOut("font attr:"+key);
//					debugOut(":"+a.getAttribute(key));
					addStyle(fontTagFont, fontTagSize, fontTagColor, null, null);
				}
			} else if (tagName.equals("PRE")) {
				// FIXME convert to style support (white-space:pre)
				preFormatted = true;
			} else if (tagName.equals("A")) {
				String namedAnchor = null;
//debugOut(getAttributeValue(a, "href"));
				Enumeration<?> atts = getAttributeEnumeration(a);
				int accessNum = -1;
				while (atts.hasMoreElements()) {
					Object key = atts.nextElement();
					if (key.toString().equalsIgnoreCase("href")) {
						String src = getAttributeValue(a, key).toString();
						//FIXME doesn't allow for possible already-set anchor via an ID
						linkSrc = getSrc(urlBase, src);
//debugOut(src+":"+linkSrc);
					} else if(key.toString().equalsIgnoreCase("name")) {
						namedAnchor = getAttributeValue(a, key).toString();
					} else if(key.toString().equalsIgnoreCase("accesskey")) {
						//TODO what about # and *? no good mapping, maybe don't bother with a bad mapping.
						try {
							accessNum = Integer.parseInt(getAttributeValue(a, key).toString());
						} catch(Exception e) {
						}
						
					}
				}
				if(accessNum >= 0 && accessNum < accessNumLinks.length) {
					accessNumLinks[accessNum] = linkSrc;
				}
				if(namedAnchor != null) {
					setAnchor(namedAnchor);
				}
			} else if (tagName.equals("BUTTON")) {
				buttonName = String.valueOf(getAttributeValue(a, "name"));
				String type = String.valueOf(getAttributeValue(a, "type"));
				// only submit if button is submit or blank type
				if(type == null || type.equals("null") || type.equalsIgnoreCase("submit") || type.trim().length() == 0) {
					buttonName += "\n"+form;
					Object val = getAttributeValue(a, "value");
					if(val != null) {
						buttonName += "\n"+String.valueOf(val);
					}
				}
			} else if (tagName.equals("SCRIPT") || tagName.equals("STYLE")
//					|| tagName.equals("NAV")
					) { //|| tagName.equals("SELECT") 
				if (ignoreUntil == null) {
					ignoreUntil = tagName;
					ignoreDepth = 1;
				}
				
//				debugOut("ignoring "+ignoreUntil);
				
			} else if(isLayingOutTables && tagName.equals("TABLE")) {
				currentTable.push( "table"+tableIndex );
				++tableIndex;
				// FIXME limbo 1 - table flag, no rows
			} else if(isLayingOutTables && tagName.equals("TR")) {
//				if(currentTable == null || currentTable.size() == 0) {
//					currentTable.push( "table"+tableIndex );
//					++tableIndex;
////				} else {
//				}
				
				int tableRowAlign = HTMLRow.ROW_ALIGN_LEFT;
				Object alignObj = getAttributeValue(a, "align"); 
				if(alignObj == null || alignObj.toString().trim().length() == 0) {
					alignObj = getStyleValue(a, "text-align");
				}
				if(alignObj != null) {
					String alignString = alignObj.toString();
					if("right".equalsIgnoreCase(alignString)) {
						tableRowAlign = HTMLRow.ROW_ALIGN_RIGHT;
					} else if("center".equalsIgnoreCase(alignString)) {
						tableRowAlign = HTMLRow.ROW_ALIGN_CENTER;
					} else if("left".equalsIgnoreCase(alignString)) {
						tableRowAlign = HTMLRow.ROW_ALIGN_LEFT;
					}
				}
				String height = (String) getAttributeValue(a, "height");
				if(height == null || height.trim().length() == 0) {
					height = getStyleValue(a, "height");
				}

				Color bgcolor = getStyleBGColor(a);
				// override style with attribute
				Object abgcolor = getAttributeValue(a, "bgcolor");
				if(abgcolor != null) {
					Color bgc = getColor(abgcolor.toString());
					if(bgc != null) {
						bgcolor = bgc;
					}
				}
				
				String bg = getStyleBGImage(a);
				
				HTMLTableRow tableRow = newTableRow(tableRowAlign, bgcolor, bg);
				tableRow.addRowHeight(height);
				
				// add row's attribute style to stack - like the body
				addStyle(null, null, null, null, null, null, bgcolor, null/*bg only comes from style*/, tableRowAlign);				
				
				//FIXME limbo 2 - table row, no cells
			} else if(isLayingOutTables && (tagName.equals("TD") || tagName.equals("TH"))) {
				newTableCell(tagName, a);
			}

		}

		private void handleFrameTag(String tagName, Object a) {
			//FRAME is usually a simple tag and won't show up here.
			//IFRAME's contents are the equivalent of "NOFRAMES" so it might not show up as a simple tag.
			String src = getAttributeString(a, "src");
			if(src != null) {
				URL url = getURL(urlBase, src);
				if(url != null) {
					src = url.toString();
				}
			String name = getAttributeString(a, "name");
			if(name == null || name.trim().length() == 0) {
				name = src;
				// limit the length of this string.
				if(name != null && name.length() > 50) {
					name = name.substring(0, 50)+"...";
				}
			}
			if (src != null && embeddedVideoPattern.matcher(src).matches()) {
				// special code for renderer to use an appropriate icon
				String poster = "@VIDEO";
				int width = 64;
				int height = 64;
				if(url != null) {
					poster = "@"+url.getHost();
					if(poster.toLowerCase().contains("youtu")) {
						URL posterUrl = YouTubeLauncher.getPoster(src);
						if(posterUrl != null) {
							width = 120;
							height = 90;
							poster = posterUrl.toString();
						}
					} else if(poster.toLowerCase().contains("vimeo")) {
						URL posterUrl = VimeoHandler.getPoster(src);
						if(posterUrl != null) {
							//FIXMe use proper sizes for vimeo poster
							width = 120;
							height = 90;
							poster = posterUrl.toString();
						}
					}
					//TODO add a VineHandler.getPoster if poster contains("vine.co") or at least a vine icon
				}
				HTMLChunk mediaChunk = new HTMLImageChunk(poster,
						width, height, src, "VIDEO");
				currentRow.add(mediaChunk);
				// save the first autoplay embedded Video or Audio frame for automatic display
				if (embeddedStream == null && src.indexOf("autoplay=1") > 0) {
					this.embeddedStream = src;
				}
			} else {
				currentRow.add(newTextChunk(tagName+": "+name, src));
			}
			}
		}

		private int sourceTagTypeTest = 0;
		private HTMLChunk sourceTagChunk = null;
		private static final int SOURCEVIDEO = 2;
		private static final int SOURCEVIDEOAUTOPLAY = 3;
		private static final int SOURCEAUDIO = 4;
		private static final int SOURCEAUDIOAUTOPLAY=5;
		private static final Pattern sourceTagTypeVideo = Pattern.compile("video/(mp(4|eg2?)|x-tivo|h264).*");
		//video/mpeg2
		//video/mpeg
		//video/mp4
		//video/x-tivo-raw-tts
		//video/x-tivo-mpeg
		private static final Pattern sourceTagTypeAudio = Pattern.compile("(audio/(x\\-)?(scm3u|mpeg|scpls)|application/x\\-winamp\\-playlist).*");
		//"audio/mpeg"
		//"audio/x-mpegurl","audio/mpeg-url","application/x-winamp-playlist"
		//"audio/scm3u","audio/x-scm3u"
		//"audio/scpls","audio/x-scpls"
		
		private void handleVideoTag(Object a) {
			String tag = "VIDEO";
			int type = SOURCEVIDEO;
			int typeautoplay = SOURCEVIDEOAUTOPLAY;
			handleVideoOrAudioTag(a, tag, type, typeautoplay);
		}

		private void handleAudioTag(Object a) {
			String tag = "AUDIO";
			int type = SOURCEAUDIO;
			int typeautoplay = SOURCEAUDIOAUTOPLAY;
			handleVideoOrAudioTag(a, tag, type, typeautoplay);
		}

		private void handleVideoOrAudioTag(Object a, String tag, int type, int typeautoplay) {
			// poster, width, height specifies image placeholder
			// only consider it embeddedStream if autoplay is present 
			//   set sourcehandling to matcher for types or to flag value if no autoplay
			//source tag within VIDEO makes VIDEO: link 
			//    if(sourcehandling) add link with type, if type matches sourcehandling, set embeddedStream
			// drop text content within VIDEO tag. 
			//    (handleText: if(sourcehandling) return
			// /VIDEO sets embeddedStream with best source entry found (if autoplay was set)
			//    clear sourcehandling
			// SOURCE: media attribute is a media query, src and type
			Enumeration<?> atts = getAttributeEnumeration(a);
			Integer width = -1;
			Integer height = -1;
			String poster = null;
			String src = null;
			String alt = null;
			while (atts.hasMoreElements()) {
				Object key = atts.nextElement();
				if (key.toString().equalsIgnoreCase("src")) {
					src = getAttributeValue(a, key).toString();
					src = getSrc(urlBase, src);
				} else if (key.toString().equalsIgnoreCase("poster")) {
					poster = getAttributeValue(a, key).toString();
				} else if (key.toString().equalsIgnoreCase("width")) {
					try {
						width = new Integer(getAttributeValue(a, key).toString());
					} catch (NumberFormatException e) {
					}
				} else if (key.toString().equalsIgnoreCase("height")) {
					try {
						height = new Integer(getAttributeValue(a, key).toString());
					} catch (NumberFormatException e) {
					}
				} else if (key.toString().equalsIgnoreCase("alt")) {
					alt = getAttributeValue(a, key).toString();
				}
			}
			if(width < 0) {
				width = getStyleSize(a, "width");
			}
			if(height < 0) {
				height = getStyleSize(a, "height");
			}
			
			// save the first embedded Video or Audio frame for automatic display
			boolean autoplay = ((Attributes) a).hasKey("autoplay");
			if (autoplay && embeddedStream == null && src != null) {//FIXME make this a list of streams
				if (embeddedVideoPattern.matcher(src).matches()) {
					this.embeddedStream = src;
				}
			}
			//   set sourcehandling to matcher for types or to flag value if no autoplay
			//source tag within VIDEO makes VIDEO: link 
			//    if(sourcehandling) add link with type, if type matches sourcehandling, set embeddedStream
			// drop text content within VIDEO tag. 
			//    (handleText: if(sourcehandling) return
			// /VIDEO sets embeddedStream with best source entry found (if autoplay was set)
			//    clear sourcehandling
			// SOURCE: media attribute is a media query, src and type
			
			if(poster == null || poster.trim().length() == 0) {
				// special code for renderer to use an appropriate icon
				poster = "@"+tag;
				if (width < 0 || height < 0) {
					width = 64;
					height = 64;
				}
				//TODO force all 2nd an later uses to use the icon
				//TODO force all IFRAMEs known to be videos to act like a VIDEO tag
			}
			HTMLChunk mediaChunk = newMediaChunk(a, tag, width, height, poster, src, alt);
			// do we need to get data from SOURCE tags?
			if(src != null && src.trim().length() > 0) {
				// set up for any SOURCE tags.
				HTMLChunk mediaChunk2 = newMediaChunk(mediaChunk);
				updateMediaChunkText(mediaChunk, null, src);
				currentRow.add(mediaChunk);
				mediaChunk = mediaChunk2;
			}
//			if (src == null || src.trim().length() == 0) {
				sourceTagChunk = mediaChunk;
				if (autoplay) {
					sourceTagTypeTest = typeautoplay;
				} else {
					sourceTagTypeTest = type;
				}
// instead of stopping at src attribute, allow possible SOURCE tags, too.
//			} else {
//				sourceTagTypeTest = 0;
//				ignoreUntil = tag;
//			}
		}

		private HTMLChunk newMediaChunk(Object a, String tag, Integer width,
				Integer height, String poster, String src, String alt) {
			HTMLChunk mediaChunk;
			if(poster == null || poster.trim().length() == 0) {
				String name = getAttributeString(a, "name");
//				if(name == null || name.trim().length() == 0) {
//					name = src;
//				}
				mediaChunk = newTextChunk((name==null || name.trim().length() == 0 ? tag : tag+": "+name), src);
			} else {
				mediaChunk = new HTMLImageChunk(poster,
						width, height, src, alt);
			}
			return mediaChunk;
		}
		
		private HTMLChunk newMediaChunk(HTMLChunk chunk) {
			if(chunk instanceof HTMLImageChunk) {
				HTMLImageChunk imageChunk = (HTMLImageChunk) chunk;
				return new HTMLImageChunk(imageChunk.src, imageChunk.width, imageChunk.height, imageChunk.link, imageChunk.alt);
			} else if(chunk instanceof HTMLTextChunk) {
				HTMLTextChunk textChunk = (HTMLTextChunk) chunk;
				return newTextChunk(textChunk.text, textChunk.link);
			}
			return newTextChunk("MEDIA",null);
		}
		
		private void handleSourceTag(Object a) {
			String type = getAttributeString(a, "type");
			if(type != null && type.trim().length() > 0) {
//				debugOut("**media source type: "+type);
				if(sourceTagTypeTest == SOURCEVIDEO || sourceTagTypeTest == SOURCEVIDEOAUTOPLAY) {
					if(!sourceTagTypeVideo.matcher(type).matches()) {
						return;
					}
				}
				else
				if(sourceTagTypeTest == SOURCEAUDIO || sourceTagTypeTest == SOURCEAUDIOAUTOPLAY) {
					if(!sourceTagTypeAudio.matcher(type).matches()) {
						return;
					}
				}
				else
					return;
			}
			// set up a new chunk for this specific SOURCE tag
			HTMLChunk mediaChunk = newMediaChunk(sourceTagChunk);
			
			//TODO media query handling
			String src = getAttributeString(a, "src");
			src = getSrc(urlBase, src);
			if(sourceTagChunk == null || src == null || src.trim().length() == 0)
				return;
			if(mediaChunk instanceof HTMLImageChunk) {
				((HTMLImageChunk)mediaChunk).link = src;
			} else if(mediaChunk instanceof HTMLTextChunk) {
				((HTMLTextChunk)mediaChunk).link = src;
				updateMediaChunkText(mediaChunk, type, src);
			}
			
			currentRow.add(mediaChunk);
			
			if (sourceTagTypeTest == SOURCEVIDEOAUTOPLAY || sourceTagTypeTest == SOURCEAUDIOAUTOPLAY) {
				//TODO handle this as a list.
				if (embeddedStream == null && src != null) {
					if (embeddedVideoPattern.matcher(src).matches()) {
						this.embeddedStream = src;
					}
				}
			}
			
//			//technically we could have multiple "good" SOURCE entries, but we only have one sourceTagChunk at the moment to work with.
//			if(sourceTagTypeTest == SOURCEVIDEO || sourceTagTypeTest == SOURCEVIDEOAUTOPLAY) {
//				ignoreUntil = "VIDEO";
//			} else if(sourceTagTypeTest == SOURCEAUDIO || sourceTagTypeTest == SOURCEAUDIOAUTOPLAY) {
//				ignoreUntil = "AUDIO";
//			}
//			sourceTagChunk = null;
//			sourceTagTypeTest = 0;
		}

		private void updateMediaChunkText(HTMLChunk mediaChunk, String type, String src) {
			if(!(mediaChunk instanceof HTMLTextChunk)) return;
			String txt = ((HTMLTextChunk)mediaChunk).text;
			if (txt.indexOf(':') < 0) {
				if (type != null && type.trim().length() > 0) {
					txt += ": " + type;
				} else if (src != null && src.trim().length() > 0) {
					txt += ": " + src;
				}
				((HTMLTextChunk) mediaChunk).text = txt;
			}
		}

		private void addListBullet() {
//				if(inList) {
					//FIXME can be set on the UL...
					// set inList to img or bulletImage or none and use it
					// but let it be overridden on the LI? or can that not happen?
					String listStyle = getListStyle();
//					debugOut("LI bullet list: "+listStyle);
					//TODO if support other types, include type="disc", "A" (value=, start=) etc.
					// skip the bullet if its style is none!
					if(listStyle == null || !listStyle.trim().toLowerCase().startsWith("none")) {
						if (listStyle != null && listStyle.trim().indexOf(":") > 0) {
							debugOut("LI bullet url: "+listStyle);
							currentRow.add(new HTMLImageChunk(
									listStyle,
									16, 16, null, null));
						} else if(listStyle != null && listStyle.trim().equalsIgnoreCase("decimal")){
							//FIXME do real numerics (and other types like alpha)
							currentRow.add(newTextChunk(" # ", null));
						} else if(listStyle != null && listStyle.trim().equalsIgnoreCase("circle")){
							if(circleImage.startsWith("http")) {
								currentRow.add(new HTMLImageChunk(
										circleImage,
										16, 16, null, null));
							}
							else {
								currentRow.add(newTextChunk(circleImage, null));
							}
						} else if(listStyle != null && listStyle.trim().equalsIgnoreCase("square")){
							if(squareImage.startsWith("http")) {
								currentRow.add(new HTMLImageChunk(
										squareImage,
										16, 16, null, null));
							}
							else {
								currentRow.add(newTextChunk(squareImage, null));
							}
						} else {
							if(bulletImage.startsWith("http")) {
								currentRow.add(new HTMLImageChunk(
										bulletImage,
										16, 16, null, null));
							}
							else {
								currentRow.add(newTextChunk(bulletImage, null));
							}
						}
					} 
					//TODO add a space after bullet?
//				}
		}
		
		private final Deque<Object> cascadingStyle = new LinkedList<Object>();
		private Styles bottomStyle = null;

		public URL baseURL;
		
		private void pushStyle(String style, boolean block) {
			if(style == null) {
				cascadingStyle.push(null);
				return;
			}
			Color color = getStyleColor(style);
			Boolean bold = getStyleFontBold(style);
			Boolean italic = getStyleFontItalic(style);
			Integer size = getStyleFontSize(style);
			String font = getStyleFontName(style);
			String list = getStyleList(style);
			Color bgcolor = getStyleBGColor(style);
			String bg = getStyleBGImage(style);
			Integer bgrepeat = getStyleBGRepeat(style);
			Integer bgposition = getStyleBGPosition(style);
			Integer align = getStyleAlign(style);
			pushStyle(color, bold, italic, size, font, list, bgcolor, bg, bgrepeat, bgposition, align, block);
		}

		private Integer getStyleAlign(String style) {
			String alignString = getStyleValue(style, "text-align");
			if(alignString != null) {
				alignString = alignString.trim();
				if("right".equalsIgnoreCase(alignString)) {
					return HTMLRow.ROW_ALIGN_RIGHT;
				} else if("center".equalsIgnoreCase(alignString)) {
					return HTMLRow.ROW_ALIGN_CENTER;
				} else if("left".equalsIgnoreCase(alignString)) {
					return HTMLRow.ROW_ALIGN_LEFT;
				}
			}
			return null;
		}

		private Integer getStyleBGPosition(String style) {
			String[] bgRepeats = getStyleValues(style, new String[] {"background", "background-position"});
			for(String bg : bgRepeats) {
				if(bg == null) continue;
				int vertical = -1;
				int horizontal = -1;
				String[] bgparts = splitStyleValues(bg);
				for(String positionString : bgparts) {
					positionString = positionString.trim().toLowerCase();
					if(positionString.equals("top")) {
						vertical = 0;
						continue;
					}
					if(positionString.equals("bottom")) {
						vertical = 100;
						continue;
					}
					if(positionString.equals("left")) {
						horizontal = 0;
						continue;
					}
					if(positionString.equals("right")) {
						horizontal = 100;
						continue;
					}
					if(horizontal < 0) {
						if(positionString.equals("center") || positionString.equals("50%")) {
							horizontal = 50;
						} else if(positionString.endsWith("%")) {
							try {
								horizontal = Integer.parseInt(positionString.substring(0, positionString.length()-1));
							} catch (Exception ex) {
							}
						}
					} else
					if(vertical < 0) {
						if(positionString.equals("center") || positionString.equals("50%")) {
							vertical = 50;
						} else if(positionString.endsWith("%")) {
							try {
								vertical = Integer.parseInt(positionString.substring(0, positionString.length()-1));
							} catch (Exception ex) {
								// oh well...
							}
						}
					}
				}
				if(horizontal >= 0) {
					if(vertical < 0) {
						// if 2nd value wasn't specified, it defaults to center.
						vertical = 50;
					}
					return horizontal*1000+vertical;
				}
			}
			return null;
		}

		private Integer getStyleBGRepeat(String style) {
			String[] bgRepeats = getStyleValues(style, new String[] {"background", "background-repeat"});
			for(String bg : bgRepeats) {
				if(bg == null) continue;
				String[] bgparts = splitStyleValues(bg);
				for(String repeatString : bgparts) {

					repeatString = repeatString.trim().toLowerCase();
					if(repeatString.contains("repeat-x")) {
						return HTMLRow.ROW_BG_REPEAT_X;
					}
					else if(repeatString.contains("repeat-y")) {
						return HTMLRow.ROW_BG_REPEAT_Y;
					}
					else if(repeatString.contains("no-repeat")) {
						return HTMLRow.ROW_BG_REPEAT_NONE;
					}
					else if(repeatString.contains("repeat")) {
						return HTMLRow.ROW_BG_REPEAT_BOTH;
					}
				}
			}
			return null;
		}

		private void pushStyle(Color color, Boolean bold, Boolean italic,
				Integer size, String font, String list, Color bgcolor, String bg, Integer bgrepeat, Integer bgposition, Integer align, boolean block) {
			Styles styleObject = new Styles(color, bold, italic, size, font, list, bottomStyle, bgcolor, bg, bgrepeat, bgposition, align, block);
			bottomStyle = styleObject;
			cascadingStyle.push(styleObject);
		}

		/**
		 * if the latest pushed style was empty, replace it with this more useful
		 * information. otherwise add to it where it was null.  Every tag pushes style, often empty, certain tags have
		 * another way to add style (default formatting, alternate attributes).
		 */
		private void addStyle(String textFont2,
				Integer textSize2, Color textColor2, Boolean isItalic, Boolean isBold) {
			addStyle(textFont2, textSize2, textColor2, isItalic, isBold, null, null, null, null);
		}
		private void addStyle(String textFont2,
				Integer textSize2, Color textColor2, Boolean isItalic, Boolean isBold, String listStyle, Color bgcolor, String bg, Integer align) {
			if(cascadingStyle.peek() == null) {
//				cascadingStyle.pop();//FIXME why was this here? doc'd that it throws exception if stack is empty!
				pushStyle(textColor2, isBold, isItalic, textSize2, textFont2, listStyle, bgcolor, bg, 
//						bgrepeat, bgposition //FIXME pass these in?
						null,null,
						align, true);
			} else if(cascadingStyle.peek() instanceof Styles){
				Styles style = (Styles)cascadingStyle.peek();
				if(isBold != null)
					style.bold = isBold;
				if(isItalic != null)
					style.italic = isItalic;
				if(textColor2 != null)
					style.color = textColor2;
				if(bgcolor != null)
					style.bgcolor = bgcolor;
				if(bg != null) {
					style.bg = bg;
//					style.bgposition = bgposition;
//					style.bgrepeat = bgrepeat;
				}
				if(textSize2 != null)
					style.size = textSize2;
				if(textFont2 != null)
					style.font = textFont2;
				if(listStyle != null)
					style.list = listStyle;
			} else {
				debugOut("COULDN'T REPLACE STYLE - existing style: "+cascadingStyle.peek());
			}
		}

		private void popStyle() {
			if(cascadingStyle.size() > 0) {
				Object poppedStyle = cascadingStyle.pop();
				if(poppedStyle != null) {
					// if the pop was something significant, it should have been the
					// bottom-most style, so revert current style to its parent
					if(poppedStyle == bottomStyle) {
						bottomStyle = bottomStyle.parent;
					} else {
						debugOut("WARNING - cascading style failed sanity check - I lost my place");
					}
				}
			}
		}

		private Object getTextFont() {
			if(bottomStyle != null)
				return bottomStyle.getFont();
			return null;
		}

		private int getTextSize() {
			if(bottomStyle != null) {
				Integer size = bottomStyle.getSize();
				if(size != null)
					return size;
			}
			return DEFAULT_TEXT_SIZE;
		}

		private boolean getBold() {
			if(bottomStyle != null) {
				Boolean bold = bottomStyle.getBold();
				if(bold != null)
					return bold;
				return false;
			}
			return false;
		}

		private String getListStyle() {
			if(bottomStyle != null) {
				return bottomStyle.getListStyle();
			}
			return null;
		}

		private boolean getItalic() {
			if(bottomStyle != null) {
				Boolean italic = bottomStyle.getItalic();
				if(italic != null)
					return italic;
				return false;
			}
			return false;
		}
		
		private Color getTextColor() {
			if(bottomStyle != null) {
				Color color = bottomStyle.getColor();
				if(color != null)
					 return color;
				return null;
			}
			return fgColor;
		}
		
		private Color getBGColor() {
			if(bottomStyle != null) {
				Color color = bottomStyle.getBGColor();
				if(color != null)
					 return color;
				return null;
			}
			return null;
		}
		
		private String getBG() {
			if(bottomStyle != null) {
				String bg = bottomStyle.getBG();
				if(bg!= null)
					 return bg;
				return null;
			}
			return null;
		}
		
		private Color getBlockBGColor() {
			if(bottomStyle != null) {
				Color color = bottomStyle.getBlockBGColor();
				if(color != null)
					 return color;
				return null;
			}
			return null;
		}
		
		private int getBlockAlign() {
			if(bottomStyle != null) {
				Integer color = bottomStyle.getBlockAlign();
				if(color != null)
					 return color;
				return HTMLRow.ROW_ALIGN_LEFT;
			}
			return HTMLRow.ROW_ALIGN_LEFT;
		}
		
		private int getBlockBGRepeat() {
			if(bottomStyle != null) {
				Integer repeat = bottomStyle.getBlockBGRepeat();
				if(repeat != null)
					 return repeat;
				return HTMLRow.ROW_BG_REPEAT_DEFAULT;
			}
			return HTMLRow.ROW_BG_REPEAT_DEFAULT;
		}
		
		private int getBlockBGPosition() {
			if(bottomStyle != null) {
				Integer position = bottomStyle.getBlockBGPosition();
				if(position != null)
					 return position;
				return HTMLRow.ROW_BG_POSITION_DEFAULT;
			}
			return HTMLRow.ROW_BG_POSITION_DEFAULT;
		}
		
		private int getBGRepeat() {
			if(bottomStyle != null) {
				Integer repeat = bottomStyle.getBGRepeat();
				if(repeat != null)
					 return repeat;
				return HTMLRow.ROW_BG_REPEAT_DEFAULT;
			}
			return HTMLRow.ROW_BG_REPEAT_DEFAULT;
		}
		
		private int getBGPosition() {
			if(bottomStyle != null) {
				Integer position = bottomStyle.getBGPosition();
				if(position != null)
					 return position;
				return HTMLRow.ROW_BG_POSITION_DEFAULT;
			}
			return HTMLRow.ROW_BG_POSITION_DEFAULT;
		}
		
		private String getBlockBG() {
			if(bottomStyle != null) {
				String bg = bottomStyle.getBlockBG();
				if(bg!= null)
					 return bg;
				return null;
			}
			return null;
		}
		
		private boolean isLocalColor() {
			if(cascadingStyle.peek() != null) {
				return ((Styles)cascadingStyle.peek()).color != null;
			}
			return false;
		}

		private String getStyleList(Object style) {
			String img = null;
			String[] bgs = getStyleValues(style, new String[] {"list-style", "list-style-image"});
			if(bgs != null && bgs.length > 0) {
				for(String bg : bgs) {
					if(bg == null) continue;
					String[] bgparts = splitStyleValues(bg);
					for(String b : bgparts) {
						String bgi = getCSSURI(b);
						if(bgi != null) {
							img = bgi;
							debugOut("list-style: url "+img);
							return getSrc(urlBase, img).toString();
						}
					}
				}
			}
			
			String[] listStyles = getStyleValues(style, new String[] {"list-style", "list-style-type"});
			String type = null;
			boolean gotNone = false;
			for(String listStyleGroup : listStyles) {
				if(listStyleGroup == null) continue;
				for(String listStyle : splitStyleValues(listStyleGroup)) {
					// not type: inside, outside, url.  possible type: none, inherit. def. type: all others
//				List<String>listStyleTypes = Arrays.asList(new String[] {
//				"disc", "circle","square", 
//				"decimal","decimal-leading-zero","lower-roman","upper-roman",
//				"lower-alpha","upper-alpha",
//				"lower-greek","lower-latin","upper-latin","armenian","georgian"});
					// go through the list... note if we see "none" and keep the definite type value
					// stop at definite type value... if we don't see it but got "none" stop there, too

					// skip the bullet if its style is none!
					if(listStyle.equals("none")) {
						gotNone = true;
						//					return "none";
					} else if(listStyle.startsWith("url")) {
					} else if(listStyle.equals("inherit")) {
					} else if(listStyle.equals("inside")) {
					} else if(listStyle.equals("outside")) {
					} else if(listStyle.equals("disc")) {
						type = "disc";
					} else if(listStyle.equals("circle")) {
						type = "circle";
					} else if(listStyle.equals("square")) {
						type = "square";
					} else {
						//FIXME just using "decimal" for all ordered list types for now.
						type = "decimal";
					}
				}
				if(type != null) {
					return type;
				} else if(gotNone) {
					return "none";
				}

			}
			return type; // probably null.
		}

		private String[] getStyleValues(Object astyle, String[] strings) {
			//FIXME this assumes arrays are length 2 always.
			
			String[] result = new String[strings.length];
			int mostSpecific = -1;
			if(!(astyle instanceof String)) {
				astyle = getAttributeValue(astyle, "style");
			}
			List<String> styleNames = Arrays.asList(strings.clone());
			for(int i = 0 ; i < styleNames.size() ; ++i) {
				styleNames.set(i, styleNames.get(i).trim().toLowerCase());
			}
				
				if(astyle != null ) {
					String styles[] = astyle.toString().split(";");
					for(String style : styles) {
						int sep = style.indexOf(":");
						if(sep > 0) {
							String name = style.substring(0, sep).trim().toLowerCase();
							int nameIdx = styleNames.indexOf( name );
							// if this is one of our names...
							if(nameIdx >= 0) {
								// ... save the value as the most specific one for this name's entry.
								String value = style.substring(sep + 1).trim();
								// if it's important, remove this name so we don't overwrite it with something later.
								if(value.endsWith("!important")) {
									int lessImp = value.length()-"!important".length();
									value = value.substring(0, lessImp).trim();
									result[nameIdx] = value;
									mostSpecific = nameIdx;
									styleNames.set(nameIdx, null);
								} else {
									result[nameIdx] = value;
									mostSpecific = nameIdx;
								}
							}
						}
					}
				}
				if(mostSpecific == 1) {
					return new String[] {result[1], result[0]};
				}
			return result;
		}

		private String getStyleBGImage(Object atts) {
			String[] bgImages = getStyleValues(atts, new String[] {"background", "background-image"});
			for(String bg : bgImages) {
				if(bg == null) continue;
				String bgi;
				String[] bgparts = splitStyleValues(bg);
				for(String b : bgparts) {
					bgi = getCSSURI(b);
					if(bgi != null) {
						return bgi;
					}
				}
			}
			return null;
		}

		private Color getStyleBGColor(Object atts) {
			String[] bgColors = getStyleValues(atts, new String[] {"background", "background-color"});
			for(String bg : bgColors) {
				if(bg == null) continue;
				Color bgc;
				String[] bgparts = splitStyleValues(bg);
				for(String b : bgparts) {
					bgc = getColor(b);
					if(bgc != null) {
						return bgc;
					}
				}
			}
			return null;
		}
		
		private Color getStyleColor(Object atts) {
			String bc = getStyleValue(atts, "color");
			Color bgc = getColor(bc);
			Color color = null;
			if(bgc != null)
				color = bgc;
			return color;
		}

		/** http://www.w3.org/TR/CSS2/fonts.html#font-boldness */
		private Boolean getStyleFontBold(Object atts) {
			String[] fontStyles = getStyleValues(atts, new String[] {"font", "font-weight"});
			for(String fontStyle : fontStyles) {
				if(fontStyle == null) continue;
				String[] parts = splitStyleValues(fontStyle);
				for(String bold : parts) {
					bold = bold.trim().toLowerCase();
					//TODO should probably handle relative values better
					if(bold.equals("bold")||bold.equals("bolder")) {
						return Boolean.TRUE;
					} else if(bold.equals("normal") || bold.equals("lighter")) {
						return Boolean.FALSE;
					}
					try {
						int weight = Integer.parseInt(bold);
						if(weight > 500) // normal is 400, bold is 700, we have to pick a dividing line
							return Boolean.TRUE;
						else
							return Boolean.FALSE;
					} catch(Exception e) {
					}
				}
			}
			return null;
		}
		
		/**
		 * Divide a css property's value into its component pieces - especially for shorthand properties like "background:" and "font:"
		 * @param styles
		 * @return
		 */
		public static String[] splitStyleValues(String styles) {
			if(splitStyleValuesPattern == null) {
				// quotes around non-quotes and \" sequences
				String quotedString = "\"(?:[^\"]|\\\")*\"";
				String tickedString = "\'(?:[^\']|\\\')*\'";
				String nonquotedString = "[^\'\"\\s\\(\\)]+";
				String string = "(?:"+quotedString+"|"+tickedString+"|"+nonquotedString+")";
				String list = "(?:" + string + "\\s*,\\s*)*" + string;
				// funcs like url('blah') rgb(0, 0, 0)
				String func = "\\w+\\(\\s*" + list + "\\s*\\)";
				// a degenerate list is just a string, so those are covered here, too
				String styleEntry = "(" + func+ "|"+ list+ ")";
				// multiple entries handled by the while loop using this same pattern
				String styleEntries = "\\s*"+styleEntry+"\\s*";

				splitStyleValuesPattern = Pattern.compile(styleEntries);
			}
			Matcher styleMatcher = splitStyleValuesPattern.matcher(styles);

			ArrayList<String> results = new ArrayList<String>();
			while(styleMatcher.find()) {
				for(int i = 1 ; i <= styleMatcher.groupCount() ; ++i) {
					String group = styleMatcher.group(i);
					if(group != null)
						results.add(group);
				}
			}
			
			return (String[]) results.toArray(new String[results.size()]);
		}

		/** http://www.w3.org/TR/CSS2/fonts.html#font-styling */
		private Boolean getStyleFontItalic(Object atts) {
			String[] fontStyles = getStyleValues(atts, new String[] {"font", "font-style"});
			for(String fontStyle : fontStyles) {
				if(fontStyle == null) continue;
				String[] parts = splitStyleValues(fontStyle);
				for(String italic : parts) {

					italic = italic.toLowerCase();
					if(italic.equals("italic")||italic.equals("oblique")) {
						return Boolean.TRUE;
					} else if(italic.equals("normal")) {
						return Boolean.FALSE;
					}
				}
			}
			return null;
		}
		
		private String getStyleFontName(Object atts) {
			String[] bgRepeats = getStyleValues(atts, new String[] {"font", "font-family"});
			for(String family : bgRepeats) {
				if(family == null) continue;
				// treat a comma-separated list as multiple space-separated entries to help us handle font-family lists a little better.
				family = family.replace(',', ' ');
				String families[] = splitStyleValues(family.toLowerCase());
//				'serif' (e.g., Times)
//				'sans-serif' (e.g., Helvetica)
//				'cursive' (e.g., Zapf-Chancery)
//				'fantasy' (e.g., Western)
//				'monospace' (e.g., Courier)
				String defaultFam = null;
				for(String fam : families) {
					fam = fam.trim();
					if(fam.endsWith(",")) {
						fam = fam.substring(0,fam.length()-1).trim();
					}
					if(fam.startsWith(",")) {
						fam = fam.substring(1).trim();
					}
					if(defaultFam == null && fam.length() > 0) {
						defaultFam = fam;
					}
					// just return generic families
					// TODO should return first one if there are no generics
					if (fam.equals("serif") || fam.equals("sans-serif")
							|| fam.equals("cursive") || fam.equals("fantasy")
							|| fam.equals("monospace")) {
						return fam;
					}
				}
				return defaultFam;
			}
			return null;
		}
		
		/** http://www.w3.org/TR/CSS2/fonts.html#font-size-props */
		private Integer getStyleFontSize(Object atts) {
			String[] bgRepeats = getStyleValues(atts, new String[] {"font", "font-size"});
			for(String bg : bgRepeats) {
				if(bg == null) continue;
				String[] bgparts = splitStyleValues(bg);
				for(String fontSize : bgparts) {
			
				fontSize = fontSize.toLowerCase();
				if (fontSize.equals("xx-small")) { // font size 1
					return 10;
				} else if (fontSize.equals("x-small")) {// between font size 1 and 2
					return 11;
				} else if (fontSize.equals("small")) {// font size 2
					return 13;
				} else if (fontSize.equals("medium")) {// font size 3, normal
					return DEFAULT_TEXT_SIZE;
				} else if (fontSize.equals("large")) {// font size 4
					return 17;
				} else if (fontSize.equals("x-large")) {// font size 5
					return 22;
				} else if (fontSize.equals("xx-large")) {// font size 6
					return 26;
				} else if(fontSize.equals("larger")) {
					return null;//not supported
				} else if(fontSize.equals("smaller")) {
					return null;//not supported
				}
				if(fontSize.indexOf('/')>0) {
					// fontSize/lineHeight
					fontSize = fontSize.substring(0, fontSize.indexOf('/'));
				}
				Integer calculatedSize = getCSSLength(fontSize);
				if(calculatedSize != null && calculatedSize > 0)
					return calculatedSize;
				}
			}
			return null;
		}

		/**
		 * currently doesn't support relative values (%, ex, em)
		 * @param styleSize
		 * @return null if length couldn't be calculated
		 */
		public static Integer getCSSLength(String styleSize) {
			// values relative to parent size
			// %, ex, em, 
			// absolute values
			// in (72pt), mm (1828.8pt), cm (182.88pt), pt, pc (12pt), px (0.75pt)
			double multiplier = 1.0;
			if(styleSize.endsWith("in")) {
				multiplier=72;
			} else if(styleSize.endsWith("mm")) {
				multiplier=1828.8;
			} else if(styleSize.endsWith("cm")) {
				multiplier=182.88;
			} else if(styleSize.endsWith("pt")) {
				multiplier=1.0;
			} else if(styleSize.endsWith("pc")) {
				multiplier=12;
			} else if(styleSize.endsWith("px")) {
//					multiplier=0.75;
				multiplier=1.0;
			} else {
				// other units not supported.
				multiplier = -1;
			}
			Integer calculatedSize = null;
			if(multiplier > 0) {
			try {
				double size = Double.parseDouble(styleSize.substring(0, styleSize.length()-2));
//debugOut("SIZE "+styleSize+" = "+size+"*"+multiplier+" = " +((int)(size*multiplier)));
				calculatedSize = new Integer((int) (size*multiplier));
			} catch(Exception e) {
			}
			}
			return calculatedSize;
		}

		/**
		 * parse CSS-style uri's of the function form "url(http://....)"
		 * @param value
		 * @return the uri from the function
		 */
		private String getCSSURI(String value) {
			if (value != null) {
				if (value.trim().toLowerCase().startsWith("url")) {
					int lp = value.indexOf("(");
					int rp = value.lastIndexOf(")");
					if (lp > 0 && rp > 0) {
						String url = value.substring(lp + 1, rp).trim();
						if(url.startsWith("\"")) {
							if(url.endsWith("\"")) {
								return url.substring(1,url.length()-1);
							}
						}
						if(url.startsWith("\'")) {
							if(url.endsWith("\'")) {
								return url.substring(1,url.length()-1);
							}
						}
						// not quoted, might have backslash-escaped characters.
						//FIXME implement un-backslashing
						return url;
					}
				}
			}
			return null;
		}

		private Color getColor(String string) {
			return ColorHelper.getColor(string);
		}

		private void newTableCell(String tagName, Object attributes) {
//			, Object rowspan,
//				Object colspan) {
			boolean isHeader = tagName.equals("TH");
			Object abgcolor = getAttributeValue(attributes, "bgcolor");
			Color bgcolor = null;
			if(abgcolor != null) {
				bgcolor = getColor(abgcolor.toString());
			}
			if(bgcolor == null) {
				Color bgc = getStyleBGColor(attributes);
				if(bgc != null)
					bgcolor = bgc;
			}
			
			Object rowspan = getAttributeValue(attributes, "rowspan"); 
			Object colspan = getAttributeValue(attributes, "colspan");
			HTMLRow htmlRow = this.tableRowStack.peek();
					//currentRow;//getRows().get(getRows().size() - 1);
			if (htmlRow instanceof HTMLTableRow) {
				HTMLTableRow row = (HTMLTableRow) htmlRow;
				Integer rowspan2 = 1;
				try {
					rowspan2 = Integer.valueOf(String.valueOf(rowspan));
				} catch (Exception e) {
				}
				Integer colspan2 = 1;
				try {
					colspan2 = Integer.valueOf(String.valueOf(colspan));
				} catch (Exception e) {
				}
				
				int align = currentRow.align;
				Object alignObj = getAttributeValue(attributes, "align"); 
				if(alignObj == null || alignObj.toString().trim().length() == 0) {
					alignObj = getStyleValue(attributes, "text-align");
				}
				if(alignObj != null) {
					String alignString = alignObj.toString();
					if("right".equalsIgnoreCase(alignString)) {
						align = HTMLRow.ROW_ALIGN_RIGHT;
					} else if("center".equalsIgnoreCase(alignString)) {
						align = HTMLRow.ROW_ALIGN_CENTER;
					} else if("left".equalsIgnoreCase(alignString)) {
						align = HTMLRow.ROW_ALIGN_LEFT;
					}
				}
				
				String width = (String) getAttributeValue(attributes, "width");
				if(width == null || width.trim().length() == 0) {
					width = getStyleValue(attributes, "width");
				}
				String height = (String) getAttributeValue(attributes, "height");
				if(height == null || height.trim().length() == 0) {
					height = getStyleValue(attributes, "height");
				}
				String bg = getBlockBG();
				if (bg != null) {
					try {
						bg = getSrc(urlBase, bg);
					} catch (Exception e) {
					}
				}
				HTMLTableCell cell = row.addCell(isHeader, rowspan2,
						colspan2, width, height, bgcolor, getBlockBGPosition(), getBlockBGRepeat(), bg);
				cell.align = align;
				currentRow = cell.rows.get(0);
//				debugOut("NEW TABLE CELL currentRow is first, empty row (already added to cell) but...rows not set?");
				
				// add cell's attribute style to stack - like the body
				addStyle(null, null, null, null, null, null, bgcolor, null/*bg only comes from style*/, align);				
			}
//			else debugOut("no NEW TABLE CELL - not a table row!");
		}

		/** extracts a named css item's value from the style attribute, if there is one */
		public static String getStyleValue(Object attributes, String styleName) {
			if(attributes instanceof String) {
				return getStyleValue((String)attributes, styleName);
			}
			Object astyle = getAttributeValue(attributes, "style");
			if(astyle == null) return null;
			return getStyleValue(astyle.toString(), styleName);
		}
		/** extracts a named css item's value from the style attribute's value */
		public static String getStyleValue(String astyle, String styleName) {
			String value = null;
			if(astyle != null && styleName != null) {
				String styles[] = astyle.toString().split(";");
				for(String style : styles) {
					int sep = style.indexOf(":");
					if(sep > 0) {
						String name = style.substring(0, sep).trim();
						if(name.equalsIgnoreCase(styleName)) {
							value = style.substring(sep + 1).trim();
							if(value.endsWith("!important")) {
								int lessImp = value.length()-"!important".length();
								value = value.substring(0, lessImp).trim();
								return value;
							}
						}
					}
				}
			}
			// returns the last entry (unless !important comes first) because that's
			// the most important how we pushed down style info
			return value;
		}

		private void setAnchor(String namedAnchor) {
			if(namedAnchor == null && linkSrc == null)
				return;
			if(linkSrc == null) linkSrc = "";
			if(namedAnchor == null) namedAnchor = "";
//else debugOut("Anchor Def: "+namedAnchor);
			int sep = linkSrc.indexOf('\t');
			if (sep >= 0) {
				// if we're replacing, remember the current one.
				if(namedAnchor.length() > 0) {
					currentRow.add(new HTMLTextChunk("", null, false,
							false, getTextSize(), null, null, -1, -1, null, linkSrc.substring(0, sep)+"\t", false, false));
				}
				linkSrc = linkSrc.substring(sep + 1);
			}
			linkSrc = namedAnchor + "\t" + linkSrc;
		}

		private void setSelected(String value) {
			// replace selected with this value.
			//FIXME multi-select has to remember all of them.
			if (select != null) {
				int end = select.indexOf("\n");
				if (end < 0)
					end = select.length();
				select = select.substring(0, select.indexOf("\t") + 1) + value
						+ select.substring(end);
			}
		}

		private void addSelectOption(String text, String value) {
			select += "\n" + text + "\t" + value;
		}

		private static Object getAttributeValue(Object a, Object key) {
			if (a instanceof MutableAttributeSet) {
				return ((MutableAttributeSet) a).getAttribute(key);
			} else if (a instanceof Attributes) {
				return ((Attributes) a).get(String.valueOf(key));
			} else if(a instanceof HashMap) {
				return ((HashMap) a).get(key);
			}
			return null;
		}

		private Enumeration<?> getAttributeEnumeration(final Object a) {
			if(a instanceof MutableAttributeSet) {
				return ((MutableAttributeSet)a).getAttributeNames();
			} else if(a instanceof Attributes) {
				return new Enumeration<String>() {
					Iterator<Attribute> iter = ((Attributes)a).iterator();
					public boolean hasMoreElements() {
						return iter.hasNext();
					}

					public String nextElement() {
						return iter.next().getKey();
					}
				};
			} else if(a instanceof HashMap) {
				debugOut("WHOOPS - calling getAttributeEnumeration on a HashMap'd attribute set!");
				return null;
			}
			return null;
		}

		@Override
		public void handleSimpleTag(Tag t, MutableAttributeSet a, int pos) {
			if(t == null) return;
			String tagName = t.toString().toUpperCase();
			boolean breaksFlow = t.breaksFlow();
			handleSimpleTag(tagName, breaksFlow, a);
		}
		public void handleSimpleTag(String tagName, boolean breaksFlow, Object a) {
			if(ignoreUntil != null) return;
			
			// adjust breaksflow Based on CSS display attribute
			String displayValue = getStyleValue(a, "display");
			// double check by display
			breaksFlow = isDisplayBlock(displayValue, breaksFlow);
			
			if(breaksFlow && inBody) {
//				debugOut("BREAK FLOW simple tag "+tagName);
				newRow();
			}
			if(tagName.equals("BASE")) {
				String href = getAttributeString(a, "href");
				baseURL = getURL(urlBase, href);
			}
			if(tagName.equals("FRAME") || tagName.equals("IFRAME")) {
				handleFrameTag(tagName, a);
			}
			if(tagName.equals("LINK")) {
				String rel = getAttributeString(a, "rel");
				if(rel != null) rel = rel.toLowerCase();
				if(rel != null && "icon".equals(rel) || "shortcut icon".equals(rel) || "apple-touch-icon".equals(rel) || "apple-touch-icon-precomposed".equals(rel)) {
//					String sizes = getAttributeString(a, "sizes");
					String href = getAttributeString(a, "href");
					//debugOut(href);
					Object type = getAttributeString(a, "type").toLowerCase();
					boolean usableType = (type != null && (type.equals("image/png")
							|| type.equals("image/gif") || type
							.equals("image/jpg")));
					if (href != null && href.trim().length() > 0
							&& (usableType || href.toUpperCase().endsWith(".PNG")
									|| href.toUpperCase().endsWith(".JPG") || href
									.toUpperCase().endsWith(".GIF"))) {
						icon = getURL(urlBase, href);
						//debugOut("icon:"+icon);
					}
				}
			} else if(tagName.equals("META")) {
				String equiv = getAttributeString(a, "http-equiv");
				if(equiv != null) equiv = equiv.toLowerCase();
				if(equiv != null && equiv.equals("refresh")) {
					this.refresh = getAttributeString(a, "content");
debugOut("META http-equiv REFRESH is "+refresh);
				}
				if(icon == null) {
					// alternate for icon
//					<meta itemprop="image" content="/rel/url.png">
					String itemprop = getAttributeString(a, "itemprop");
					if(itemprop != null && itemprop.toLowerCase().equals("image")) {
						String content = getAttributeString(a, "content");
						if(content != null && content.trim().length() > 0) {
							icon = getURL(urlBase, content);
						}
					}
				}
			} else if (tagName.equals("A")) {
				// allow for simple entry to create a named anchor location
				String namedAnchor = null;
//debugOut(getAttributeValue(a, "href"));
				Enumeration<?> atts = getAttributeEnumeration(a);
				while (atts.hasMoreElements()) {
					Object key = atts.nextElement();
					if(key.toString().equalsIgnoreCase("name")) {
						namedAnchor = getAttributeValue(a, key).toString();
					}
				}
				// can't just set anchor...never will have text that uses it.
				if(namedAnchor != null) {
					setAnchor(namedAnchor);
					currentRow.add(new HTMLTextChunk("", null, false,
							false, getTextSize(), null, null, -1, -1, null, linkSrc, false, false));
					setAnchor(null);
				}
				
			} else
			if(tagName.equals("INPUT")) {
				currentRow.add(new HTMLFormInput(form, getAttributeMap(a), getTextFont(), getBold(), getItalic(), getTextSize()));
			} else if(tagName.equals("HR")) {
				int width = -1;
				int height = -1;
				float percent = 0.0f;
				Enumeration<?> atts = getAttributeEnumeration(a);
				while (atts.hasMoreElements()) {
					Object key = atts.nextElement();
					if (key.toString().equalsIgnoreCase("width")) {
						String wid = getAttributeValue(a, key).toString();
						try {
							if (wid.trim().endsWith("%")) {
								percent = new Float(wid.trim().substring(0,
										wid.trim().length() - 2));
							} else {
								width = new Integer(wid);
							}
						} catch (NumberFormatException e) {
						}
					} else if (key.toString().equalsIgnoreCase("height")) {
						try {
							height = new Integer(getAttributeValue(a, key).toString());
						} catch (NumberFormatException e) {
						}
				
					}
				}
				if(width < 0 && percent <= 0.0f) {
					String wid = getStyleValue(a, "width");
					try {
						if (wid != null && wid.trim().endsWith("%")) {
							percent = new Float(wid.trim().substring(0,
									wid.trim().length() - 2));
						} else {
							width = getStyleSize(a, "width");
						}
					} catch (NumberFormatException e) {
					}
				}
				if(height < 0) {
					height = getStyleSize(a, "height");
				}
				currentRow.add(new HTMLHRChunk(percent, width, height));
				newRow();
			} else if (tagName.equals("IMG")) {
				String src = null;
				String alt = null;
				int width = -1;
				int height = -1;
				Enumeration<?> atts = getAttributeEnumeration(a);
				while (atts.hasMoreElements()) {
					Object key = atts.nextElement();
					if (key.toString().equalsIgnoreCase("src")) {
						src = getAttributeValue(a, key).toString();
					} else if (key.toString().equalsIgnoreCase("width")) {
						try {
							width = new Integer(getAttributeValue(a, key).toString());
						} catch (NumberFormatException e) {
						}
					} else if (key.toString().equalsIgnoreCase("height")) {
						try {
							height = new Integer(getAttributeValue(a, key).toString());
						} catch (NumberFormatException e) {
						}

					} else if (key.toString().equalsIgnoreCase("alt")) {
						alt = getAttributeValue(a, key).toString();
					}
				}
				if(width < 0) {
					width = getStyleSize(a, "width");
				}
				if(height < 0) {
					height = getStyleSize(a, "height");
				}
				if(src != null) {
					String link = linkSrc;
					if(buttonName != null) {
						link = "BUTTON:"+buttonName;
					}
					currentRow.add(new HTMLImageChunk(getSrc(urlBase, src),
							width, height, link, alt));
					// consider the anchor used up
					setAnchor(null);
				}
			} else if(tagName.equals("SOURCE")) {
				handleSourceTag(a);
			} else if(tagName.equals("VIDEO")) {
				handleVideoTag(a);
				this.sourceTagChunk = null;
				this.sourceTagTypeTest = 0;
			} else if(tagName.equals("AUDIO")) {
				handleAudioTag(a);
				this.sourceTagChunk = null;
				this.sourceTagTypeTest = 0;
			}
		}

		private int getStyleSize(Object a, String sizeStyle) {
			String wid = getStyleValue(a, sizeStyle);
			//FIXME handle more types than px, even if it doesn't include percent
			if(wid != null && wid.trim().endsWith("px")) {
				try {
					wid = wid.trim();
					wid = wid.substring(0, wid.length()-2).trim();
					return new Integer(wid);
				} catch (NumberFormatException e) {
				}
			}
			return -1;
		}

		private static void debugOut(Object output) {
			if(debuglevel == null) {
				debuglevel = Integer.getInteger("com.tivo.hme.sdk.debug.level", 0);
			}
			if(debuglevel > 0)
				System.out.println(output);
		}

		private String getAttributeString(Object a, String key) {
			Object base = getAttributeValue(a, key);
			if(base == null) return null;
			return String.valueOf(base);
		}

		private HashMap<String, String> getAttributeMap(Object at) {
			LinkedHashMap<String, String> attributes = new LinkedHashMap<String, String>();
			
			if(at instanceof MutableAttributeSet) {
				MutableAttributeSet a = (MutableAttributeSet) at;
				// attribute set doesn't survive.  save the information.
				Enumeration<?> atts = a.getAttributeNames();
				while (atts.hasMoreElements()) {
					Object key = atts.nextElement();
					Object val = a.getAttribute(key);
					attributes.put(String.valueOf(key), String.valueOf(val));
				}
			} else if(at instanceof Attributes) {
				Attributes a = (Attributes) at;
				for(Attribute attr : a) {
					attributes.put(attr.getKey(), attr.getValue());
				}
			}
			return attributes;
		}

		private String getSrc(URL urlBase, String src) {
			if(src == null) return null;
			try {
				// fix spaces in url
				src = src.trim().replaceAll(" ", "%20");
				return new URL(urlBase, src).toString();
			} catch (MalformedURLException e) {
				debugOut(e);
				return src;
			}
		}
		
		private URL getURL(URL urlBase, String src) {
			if(src == null) return null;
			try {
				src = src.trim().replaceAll(" ", "%20");
				return new URL(urlBase, src);
			} catch (MalformedURLException e) {
				debugOut(e);
				return null;
			}
		}

		@Override
		public void handleEndTag(Tag t, int pos) {
			if(t == null) return;
			String tagName = t.toString().toUpperCase();
			boolean breaksFlow = t.breaksFlow();
			handleEndTag(tagName, breaksFlow);
		}
		public void handleEndTag(String tagName, boolean breaksFlow) {
			if (ignoreUntil != null) {
				if (tagName.equals(ignoreUntil)) {
					ignoreDepth--;
					if(ignoreDepth <=0) {
						ignoreUntil = null;
					}
					return;
				}
//debugOut(ignoreUntil+" skippping "+tagName);
			}
			popStyle();
			if(tagName.equals("BODY")) {
				// assuming /body exists, it's a good chance to do cleanup.
				if(title != null && title.equals(TITLE_FLAG_VALUE)) {
					title = null;
				}
				inBody = false;
			}
//debugOut("/"+t);
//			// TODO Auto-generated method stub
//			super.handleEndTag(t, pos);
			if(breaksFlow && inBody) {
				newRow();
			}
			
			if(tagName.equals("VIDEO")||tagName.equals("AUDIO")) {
				// no longer need to watch for SOURCE tags.
				sourceTagChunk = null;
				sourceTagTypeTest = 0;
			} else if(tagName.equals("UL")) {
				inList  = false;
			} else if(tagName.equals("SELECT")) {
				String type = "select-one";
				HashMap<String, String> selectAttributes = getAttributeMap(select_attributes);
				if(selectAttributes.containsKey("multiple")) {
					type = "select-multiple";
				}

				String name = getAttributeString(select_attributes, "name");
				HTMLFormSelect selectChunk = new HTMLFormSelect(form, selectAttributes,
						name, type,
						select, getTextFont(), getBold(),
						getItalic(), getTextSize());
				currentRow.add(selectChunk);
				option_value = null;
				select = null;
				select_attributes = null;
			} else if(tagName.equals("OPTION")) {
				if(option_value != null) {
					addSelectOption("", option_value);
					option_value=null;
				}
			} else if(tagName.equals("FORM")) {
				form = null;
			} else if(tagName.equals("TEXTAREA")) {
				if(form_textarea != null) {
					// must have been an empty textarea
					HashMap<String, String> textareaAttributes = getAttributeMap(textarea_attributes);
					HTMLFormTextarea textarea = new HTMLFormTextarea(form_textarea,
							"", form, textareaAttributes, getTextFont(), getBold(), getItalic(),
							getTextSize());
					currentRow.add(textarea);
				}
				form_textarea = null;
				textarea_attributes = null;
//				preFormatted = false;// no idea why this was here.
			} else if(tagName.equals("PRE")) {
				preFormatted = false;
			} else if(tagName.equals("A")) {
//debugOut("end anchor");
				currentRow.add(new HTMLTextChunk("", null, false,
						false, getTextSize(), null, null, -1, -1, null, linkSrc, false, false));
				linkSrc = null;
			} else if (tagName.equals("BUTTON")) {
				buttonName = null;
			} else if(isLayingOutTables && tagName.equals("TABLE")) {
				if(currentTable.size() > 0)
//				currentTable.pop();
					currentTable.clear();//FIXME temporary until proper nested table support.
				newRow();
			}
		}
		
		@Override
		public void handleText(char[] data, int pos) {
			String text = new String(data);
			handleText(text, text);
		}
		public void handleText(String text, String wholeText) {
			if(ignoreUntil != null) {
//debugOut(ignoreUntil+" skipping text:"+text);
				return;
			}
			if(sourceTagChunk != null) {
				//We're in a VIDEO or AUDIO tag looking for SOURCE tags 
				// - any text content is for non-conformant browsers.
				return;
			}
//if(new String(data).indexOf("noscript") >=0) {
//debugOut("!!!"+new String(data));
//}
//			// TODO Auto-generated method stub
//			super.handleText(data, pos);
//debugOut(text);
			if (title != null && title.equals(TITLE_FLAG_VALUE)) {
				title = text;
			} else if (select != null && option_value != null) {
				addSelectOption(text, option_value);
				option_value = null;
			} else if (form_textarea != null) {
				HashMap<String, String> textareaAttributes = getAttributeMap(textarea_attributes);
				HTMLFormTextarea textarea = new HTMLFormTextarea(form_textarea,
						text, form, textareaAttributes, getTextFont(), getBold(), getItalic(),
						getTextSize());
				currentRow.add(textarea);
				// done with it.
				form_textarea = null;
				textarea_attributes = null;
			} else if (inBody) {
//debugOut(text);
				String link = linkSrc;
				if(buttonName != null) {
					//FIXME doesn't allow for button's ID as an anchor destination
					link = "BUTTON:"+buttonName;
				}
				if(preFormatted && wholeText.indexOf("\n") > -1) {
					// use the "wholeText" that still has newlines
					String[] textrows = wholeText.split("\\n");
debugOut("preformatted splitting into " + textrows.length + " rows");
//debugOut("first: "+textrows[0]);
					if(textrows.length > 0) {
						currentRow.add(newTextChunk(textrows[0], link));
					}
					
					// add all middle rows as own rows
					for(int i = 1 ; i < textrows.length-1 ; ++i) {
						String row = textrows[i];
						newRow();
//if(i < 10)
//debugOut(rows.size()+":"+row);
						currentRow.add(newTextChunk(row, link));
					}
					newRow();
					// continue with the last row
					if(textrows.length > 0) {
						text = textrows[textrows.length-1];
					} else {
						text = "";
					}
					
//debugOut("LAST: "+text);
				}
				
//				if(!(text!=null && text.trim().length() > 0)){link = null;}
				HTMLStyleChunk chunk = newTextChunk(text, link);
				// consider the anchor used up
////				if(text!=null && text.trim().length() > 0) 
//				if(link != null)
//					setAnchor(null);
//				debugOut("ADD TEXT to current row");
				currentRow.add(chunk);
			}
		}

		private HTMLTextChunk newTextChunk(String text, String link) {
			Color bgc = getBGColor();
			String bg = getBG();
			int bgr = HTMLRow.ROW_BG_REPEAT_BOTH;
			int bgp = HTMLRow.ROW_BG_POSITION_LEFT_TOP;
			if (currentRow != null) {
				if (currentRow.bgcolor != null && currentRow.bgcolor.equals(bgc)) {
					bgc = null;
				}
				//FIXME should compare absolute urls
				if (currentRow.bg != null && currentRow.bg.equals(bg)) {
					bg = null;
				}
			}
			if(bg != null) {
				bgr = getBGRepeat();
				bgp = getBGPosition();
			}
			return new HTMLTextChunk(text, getTextFont(), getBold(),
					getItalic(), getTextSize(), getTextColor(), bgc, bgp, bgr, bg, link, false, isLocalColor());
		}

		synchronized private void newRow() {
			if(currentRow instanceof HTMLTableRow) {
				// pop the last table row off the stack.
				//FIXME (This has to do with nested tables) why is this? document... this is a complicated and important method.
				if(this.tableRowStack.size() > 0)
				this.tableRowStack.pop();
			}
			if(currentTable != null && currentTable.size() > 0) {
				HTMLTableCell lastCell = null;
				// currentRow is the row of the current cell
				if(currentRow instanceof HTMLTableRow) {
					if(((HTMLTableRow)currentRow).getHTMLChunks().size() > 0) {
						// it's possible to only have a non-cell entry in the row, so there is no "last cell"
						lastCell = ((HTMLTableRow)currentRow).getLastCell();
					}
				}
				else if(tableRowStack.size() > 0) {
					// currentRow is in a table cell
					lastCell = tableRowStack.peek().getLastCell();
				}
				//otherwise, we're in limbo... a table, but now tablerows established yet
				if(lastCell == null) {
//					debugOut("no NEW ROW current Table Row has no last cell");
					// same as if gethtmlchunks.size was 0
				} else {
//					debugOut("NEW ROW in table's last cell");
					newRow(lastCell.rows);
				}

			} else {
//				debugOut("NEW ROW");
				newRow(getRows());
			}
		}

		synchronized private HTMLTableRow newTableRow(int tableRowAlign, Color bgcolor, String bg) {
			if(currentRow instanceof HTMLTableRow) {
				// pop the last cell of the previous row off the stack.
				if(this.tableRowStack.size() > 0)
				this.tableRowStack.pop();
			}
			// don't consolidate empty TableRows since they should be considered significant still.
//			HTMLRow lastRow = null;
//			HTMLRow curRow = null;
//			// end row in rows is "currentRow"
//			if(rows.size() > 1) {
//				// use temp variable because currentRow might be last row of table,
//				// while this value is the whole table row
//				curRow = rows.get(rows.size() - 1);
//			}
//			if(rows.size() > 2) {
//				lastRow = rows.get(rows.size() - 2);
//			}
//					
//			if (emptyRow(curRow) && ( 
//					( lastRow != null && emptyRow(lastRow)))) {
////debugOut("skip empty row");
//				// skip adding a new row if it's still empty and the previous
//				// row was empty.
//				// create a new one since it might just be "virtually" empty
//				currentRow = new HTMLTableRow();
//				// replace the current row in the rows list with it
//				rows.remove(rows.size()-1);
//				rows.add(currentRow);
//			} else {
//			debugOut("NEW TABLE ROW set currentRow and getRows().add it");
				currentRow = new HTMLTableRow(currentTable.peek());
				
				getRows().add(currentRow);
				this.tableRowStack.push((HTMLTableRow) currentRow);
//			}
			currentRow.align = tableRowAlign;
			currentRow.bgcolor = bgcolor;
			currentRow.bg = bg;
			
			currentlyCombiningLinks = 0;
			
//			getRows().get(getRows().size()-1).mergeChunks(); ... uhh there aren't any chunks to merge yet.
			return (HTMLTableRow) currentRow;
		}
		
		synchronized private void newRow(List<HTMLRow> rows) {
//			if(currentRow != null)debugOut("add row, last row had "+currentRow.getHTMLChunks().size());
			HTMLRow curRow = null;
			HTMLRow lastRow = null;
			// end row in rows is "currentRow"
			if(rows.size() > 1) {
				// use temp variable because currentRow might be last row of table,
				// while this value is the whole table row
				curRow = rows.get(rows.size() - 1);
			}
			if(rows.size() > 2) {
				lastRow = rows.get(rows.size() - 2);
			}
					
			if (emptyRow(curRow) && ( currentlyCombiningLinks>0 || 
					( lastRow != null && emptyRow(lastRow)))) {
//debugOut("skip empty row");
//				currentlyCombiningLinks = 0;
				// skip adding a new row if it's still empty and the previous
				// row was empty.
				// create a new one since it might just be "virtually" empty
//				debugOut("NEW ROW currentRow and replace/add");
				currentRow = new HTMLRow();
				currentRow.bgcolor = getBlockBGColor();
				currentRow.bg = getBlockBG();
				currentRow.setBGProps(getBlockBGRepeat(), getBlockBGPosition());
				currentRow.align = getBlockAlign();
				// replace the current row in the rows list with it
				rows.remove(rows.size()-1);
				rows.add(currentRow);
			} else if (groupIndividualLinks
					&& simpleLinkRow(curRow)
					// have to share style with other links we're combining
					&& (currentlyCombiningLinks == 0 || getLinkSample(
							curRow).sameBaseStyleAs(combiningLinkSample))
							// we're combining already, or aren't, but the last row is also a simple link row or blank
					&& (lastRow != null && 
					((currentlyCombiningLinks > 0) 
							|| (simpleLinkRow(lastRow) || emptyRow(lastRow))))) {
//debugOut("combine link rows");
				if(currentlyCombiningLinks == 0) {
					combiningLinkSample = getLinkSample(curRow);
				}
				if(currentlyCombiningLinks >= maxCombinedLinks) {
					currentlyCombiningLinks = 1;
				} else {
					currentlyCombiningLinks++;
					//FIXME this still isn't quite doing it... probably once combined it needs a fancier test
					if(!emptyRow(lastRow, true)) { // ignore anchor as only source of non-empty
						if(linkSeparator.startsWith("http")) {
							lastRow.add(new HTMLImageChunk(linkSeparator, 16, 16, null, null));
						} else {
							lastRow.add(new HTMLTextChunk(linkSeparator, null, true, false,
									15, Color.black, null, -1, -1, null, null, false, false));
						}
					}
					for (HTMLChunk chunk : curRow.getHTMLChunks()) {
						lastRow.add(chunk);
					}
					// replace the current row in the rows list with the new empty row
					rows.remove(rows.size()-1);
				}
				// empty out the currentRow
//				debugOut("NEW ROW currentRow and add");
				currentRow = new HTMLRow();
				currentRow.bgcolor = getBlockBGColor();
				currentRow.bg = getBlockBG();
				currentRow.align = getBlockAlign();
				rows.add(currentRow);
			} else {
//if (simpleLinkRow(currentRow)) {
//	debugOut((lastRow != null) + " - "
//			+ (lastRow != null && (currentlyCombiningLinks>0))
//			+ " - "
//			+ (lastRow != null && simpleLinkRow(lastRow)));
//}
//debugOut("regular row");
				currentlyCombiningLinks = 0;
//				debugOut("NEW ROW currentRow and add");
				currentRow = new HTMLRow();
				currentRow.bgcolor = getBlockBGColor();
				currentRow.bg = getBlockBG();
				currentRow.align = getBlockAlign();
				
				rows.add(currentRow);
			}
			
			rows.get(rows.size()-1).mergeChunks();

		}
		
		private HTMLTextChunk getLinkSample(HTMLRow row) {
			HTMLTextChunk betterThanNothing = null;
			for(HTMLChunk chunk : row.getHTMLChunks()) {
				if(chunk instanceof HTMLTextChunk) {
					HTMLTextChunk html = (HTMLTextChunk) chunk;
					if(html.link != null && html.link.trim().length() == 0) {
						return html;
					}
					betterThanNothing = html;
				}
			}
			//TODO could be more robust to not return null
			return betterThanNothing;
		}

		private boolean simpleLinkRow(HTMLRow row) {
			if(row==null || row.getHTMLChunks().size() == 0) return false;
			String gotone = null;
			for(HTMLChunk chunk : row.getHTMLChunks()) {
				if(chunk instanceof HTMLTextChunk) {
					HTMLTextChunk html = (HTMLTextChunk) chunk;
					if(html.text == null || html.text.trim().length() == 0) {
						// doesn't count, try the next one.
						continue;
					}
					if(html.link == null || html.link.trim().length() == 0) {
//if(gotone != null) debugOut("link ended");
						return false;
					}
//System.out.print(" combine "+html.text+". ");
					if(gotone != null) {
						if(html.link.equals(gotone)) {
							// still good
						} else {
//debugOut("html.link changed");
							return false;
						}
					}
					gotone = html.link;
				} else {
					return false;
				}
			}
//debugOut("ok");
			return true;
		}

		private boolean emptyRow(HTMLRow row) {
			return emptyRow(row, false);
		}
		private boolean emptyRow(HTMLRow row, boolean ignoreAnchors) {
			//not counting null as empty... works for this use
			if(row == null) return false;
			if( row.getHTMLChunks().size() == 0 )
				 return true;
			else {
				//  also test chunks for all being empty text
				for(HTMLChunk chunk : row.getHTMLChunks()) {
					if(chunk instanceof HTMLTextChunk) {
						HTMLTextChunk html = (HTMLTextChunk) chunk;
						if(html.text == null || html.text.equals(bulletImage) || (html.text.trim().length() == 0
								&& (ignoreAnchors || html.anchor == null) // FIXME ??this should be ok, now, we keep the anchor for next item if text was empty
								)) {
							// empty so far, try the next one.
							continue;
						} else {
							return false;
						}
					} else if (chunk instanceof HTMLImageChunk) {
						HTMLImageChunk image = (HTMLImageChunk) chunk;
						if(image.src != null && image.src.equals(bulletImage)) {
							// a bullet can still count as emptiness
							continue;
						} else {
							return false;
						}
					} else {
						return false;
					}
				}
				return true;
			}
		}

		
		// JSOUP
		//BEGIN NodeVisitor
		
		public void head(Node node, int depth) {
//if("www.google.com".equals(urlBase.getHost()))
//debugOut(node);
			if(node instanceof TextNode) {
				handleText(((TextNode)node).text(), ((TextNode)node).getWholeText());
			} else if(node instanceof Element) {
				org.jsoup.parser.Tag tag = ((Element)node).tag();
				String tagName = tag.getName().toUpperCase();
				boolean formatAsBlock = formatAsBlock(tagName, tag.formatAsBlock());
				if(tag.isEmpty()) {
					handleSimpleTag(tagName, formatAsBlock, node.attributes());
				} else {
					handleStartTag(tagName, formatAsBlock, node.attributes());
				}
				
			}
		}

		private boolean formatAsBlock(String tagName, boolean formatAsBlock) {
			//adjust formatAsBlock
			// we disagree - should still block format on  p, h1-h6, and li
			if (!formatAsBlock
				&& (tagName.equals("BR")
						|| tagName.equals("LI"))) {
				return true;
//			} else if (formatAsBlock
//					&& (tagName.equals("TD") || tagName.equals("TH"))) {
//				return false;
			} else {
				return formatAsBlock;
			}
		}

		public void tail(Node node, int depth) {
			if(node instanceof Element) {
				org.jsoup.parser.Tag tag = ((Element)node).tag();
				if(tag.isEmpty())
					return;
				String tagName = tag.getName().toUpperCase();
				
				boolean breaksFlow = tag.formatAsBlock();
				// adjust breaksflow Based on CSS display attribute
				String displayValue = getStyleValue(node.attributes(), "display");
				// double check by display
				breaksFlow = isDisplayBlock(displayValue, breaksFlow);
				
				boolean formatAsBlock = formatAsBlock(tagName, breaksFlow);
				handleEndTag(tagName, formatAsBlock);
			}
		}

		private boolean isDisplayBlock(String displayValue, boolean breaksFlow) {
			if(displayValue != null) {
				displayValue = displayValue.trim();
				if(displayValue.equals("inherit")){
					// TODO support inheritance
				} else if(
						displayValue.equals("none") ||
						displayValue.startsWith("inline")) {
					breaksFlow = false;
				} else if(displayValue.startsWith("table-cell")){
					// table cells aren't something to mess with.
					// FIXME if you set breaksFlow to true, tables break.
				} else {
					// all others are blocks, I believe
					breaksFlow = true;
				}
			}
			return breaksFlow;
		}

		private LinkedList<HTMLTableRow> tableRowStack = new LinkedList<HTMLTableRow>();
		public List<HTMLRow> getRows() {
//			if(tableRowStack.size() == 0)
				return rows;
//			else {
//				HTMLTableRow tableRow = tableRowStack.peek();
//				HTMLTableCell cell = tableRow.getLastCell();
//				if(cell != null)
//					return cell.rows;
//				else
//					return rows;
//			}
		}

		//END NodeVisitor
		// /JSOUP
		
	}