package com.blackledge.david.html.hme;

import java.awt.Color;
import java.awt.Rectangle;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map.Entry;
import java.util.Vector;

import com.blackledge.david.html.UnicodeHelper;
import com.blackledge.david.tivo.V49BApplication;
import com.blackledge.david.tivo.V49IHmeProtocol;
import com.blackledge.david.tivo.widgets.ActionButton;
import com.blackledge.david.tivo.widgets.BNumberedList;
import com.blackledge.david.tivo.widgets.typo.TyPo;
import com.blackledge.david.tivo.widgets.typo.TyPo.Keyboard;

import com.tivo.hme.bananas.BApplication;
import com.tivo.hme.bananas.BButton;
import com.tivo.hme.bananas.BEvent;
import com.tivo.hme.bananas.BKeyboard;
import com.tivo.hme.bananas.BScreen;
import com.tivo.hme.bananas.BView;
import com.tivo.hme.sdk.HmeEvent;
import com.tivo.hme.sdk.HmeEvent.FontInfo;
import com.tivo.hme.sdk.StreamResource;
import com.tivo.hme.sdk.View;

public class BookmarkScreen extends BScreen implements VideoBackgroundScreen {

	private BookmarkList list;
	private LinkedHashMap<String, URL[]> bookmarks;
	private int bookmarkStartIndex;
	private TyPo typo;
	private String typedURL = "http://"; //$NON-NLS-1$
	private BKeyboard bkey;
	private ActionButton cancel;
	private ActionButton ok;
	private boolean isTypo = true;

	private static void debugOut(String output) {
		WebBrowserApp.debugOut(output);
	}
	
	private static int getRowFontSize(BScreen screen) {
		float ratio = ((float) screen.getHeight()) / 480f;
		return (int) (15*(1.0+(ratio-1.0)/2));
	}

	private static class BookmarkList extends BNumberedList {

		private BookmarkScreen screen;

		public BookmarkList(BookmarkScreen screen, BView parent, int x, int y,
				int width, int height, int rowHeight) {
			super(parent, x, y, width, height, rowHeight);
			this.screen = screen;
		}

		/**
		 * Handle focus movement.
		 */
		public boolean handleFocus(boolean isGained, BView gained, BView lost) {
			// work around a bug in BList when gained is null.
			if (!isGained && gained == null) {
				refresh();
				return true;
			} else {
				return super.handleFocus(isGained, gained, lost);
			}
		}
		
		@Override
		protected void createRow(BView parent, int index) {
			// let BNumberedList create the numbering view
			super.createRow(parent, index);
			
//			debugOut("create row " + index);
			if (screen.bookmarks == null || index < size()) {
				if (index < screen.bookmarkStartIndex) {
					BButton button = new BButton(parent, 0, 0,
							parent.getWidth(), parent.getHeight());
					button.setTranslation(10, 0);
					String text = String.valueOf(get(index));
					if(text.equals(ADVERTISEMENT)) {
						createAdRow(parent, index);
					} else {
						int fontSize = this.getRowHeight();
						fontSize = getRowFontSize(getScreen())+5;
						String fontString = "default-"+(fontSize)+"-bold.font"; //$NON-NLS-1$ //$NON-NLS-2$
						restrictWidth(text, fontString, button.getWidth());
						button.setResource(button.createText(
								fontString, Color.white,
								UnicodeHelper.fixTextForRendering(text)), RSRC_HALIGN_LEFT|RSRC_VALIGN_CENTER);
					}
//debugOut("action:" + get(index)+" "+screen.bookmarkStartIndex);
				} else if (screen.bookmarks != null) {
//					debugOut("bkmrk " + get(index));
					BButton button = new BButton(parent, 0, 0,
							parent.getWidth(), parent.getHeight());
					Entry<String, URL[]> bookmark = screen.getBookmark(index
							- screen.bookmarkStartIndex);
					String title = bookmark.getKey();
					boolean isHME = bookmark.getValue()[0].getProtocol()
							.startsWith("hme"); //$NON-NLS-1$
					boolean isFolder = bookmark.getValue()[0].toString().startsWith(WebBrowserApp.BOOKMARK_FOLDER_PREFIX);
					Color color = isHME ? Color.white : Color.yellow;
					if(isFolder) {
						color = Color.cyan;
						if(title != null)
							title = title.trim() + "/";
					}
					int fontSize = this.getRowHeight();
					String fontString = "default-"+fontSize+".font"; //$NON-NLS-1$ //$NON-NLS-2$
					title = restrictWidth(title, fontString, button.getWidth());
					
					Object icon = null;
					if (bookmark.getValue().length > 1
							&& bookmark.getValue()[1] != null) {
						icon = bookmark.getValue()[1];
					} else if(isFolder) {
						String genre = null;
						// make the genre folders and their immediate children use genre folder icons
						String f = ((BookmarkScreen) getScreen()).folder;
						if (f.endsWith("Apps")) {
							if (title.equals("photos/")) {
								genre = "photos";
							} else if (title.equals("video/")) {
								genre = "video";
							} else if (title.equals("music/")) {
								genre = "music";
							}
						} else if (f.endsWith("Apps" + WebBrowserApp.BOOKMARK_FOLDER_SEPARATOR
								+ "photos")) {
							genre = "photos";
						} else if (f.endsWith("Apps" + WebBrowserApp.BOOKMARK_FOLDER_SEPARATOR
								+ "video")) {
							genre = "video";
						} else if (f.endsWith("Apps" + WebBrowserApp.BOOKMARK_FOLDER_SEPARATOR
								+ "music")) {
							genre = "music";
						}
						icon = ((WebBrowserApp)getBApp()).getTheme().getFolderIcon(genre);
					}
					if(isHME) {
						//FIXME if icon is null, and url is hme:, try to get the icon.png
					}
					
					if(icon != null) {
						int iconSize = 24;
						BView view = new BView(button, 10, (button.getHeight() - iconSize) / 2, iconSize,
								iconSize);
//TODO don't set resource if it's no good or something... simulator seems to think we're setting it
						view.setResource(icon,
								HmeEvent.RSRC_IMAGE_BESTFIT
										| HmeEvent.RSRC_HALIGN_RIGHT
										| HmeEvent.RSRC_VALIGN_CENTER);
						title = "    "+title; //$NON-NLS-1$
					}
					button.setResource(button.createText(fontString,
							color, UnicodeHelper.fixTextForRendering(title)));
				}
			} 
//			else debugOut("huh?");
		}

		private void createAdRow(BView parent, int index) {
			Object icon = getAdIcon(index);
			String title = getAdTitle(index);
			boolean isBanner = (title == null);
			
			if(isBanner && icon != null) {
				//Try to make ad banners 60px tall (or two rows at most)
				int adHeight = 60;
				adHeight = Math.min(adHeight, this.getRowHeight()*2);
				int add = 0;
				if (parent.getHeight() < adHeight) {
					add = adHeight - parent.getHeight();
				}
				parent.setBounds(parent.getX(), parent.getY() - (add/2),
						parent.getWidth(), parent.getHeight() + add);
			}
			
			BButton button = new BButton(parent, 0, 0,
					parent.getWidth(), parent.getHeight());
			
			if(isBanner && icon != null) {
				button.setResource(icon,
						HmeEvent.RSRC_IMAGE_BESTFIT
						| HmeEvent.RSRC_HALIGN_CENTER
						| HmeEvent.RSRC_VALIGN_TOP);
			} else {
				Color color = Color.orange;
				int fontSize = this.getRowHeight();
				String fontString = "default-"+fontSize+"-bolditalic.font"; //$NON-NLS-1$ //$NON-NLS-2$
				title = restrictWidth(title, fontString, button.getWidth());

	//			if(isHME) {
	//				//FIXME if icon is null, and url is hme:, try to get the icon.png
	//			}

				if(icon != null) {
					int iconSize = 24;
					BView view = new BView(button, 10, (button.getHeight() - iconSize) / 2, iconSize,
							iconSize);
					//TODO don't set resource if it's no good or something... simulator seems to think we're setting it
					view.setResource(icon,
							HmeEvent.RSRC_IMAGE_BESTFIT
							| HmeEvent.RSRC_HALIGN_RIGHT
							| HmeEvent.RSRC_VALIGN_CENTER);
					title = "    "+title; //$NON-NLS-1$
				}
				button.setResource(button.createText(fontString,
						color, UnicodeHelper.fixTextForRendering(title)));
			}
		}

		private HashMap<Integer, Object[]> ads = new HashMap<Integer, Object[]>();
		private Object[] getAd(int index) {
			Object[] ad ;
			synchronized(ads) {
				ad = ads.get(index);
				if(ad == null) {
					ad = WebBrowserApp.getRandomAd();
					ads.put(index, ad);
				}
			}
			return ad;
		}
		
		private String getAdTitle(int index) {
			Object[] ad = getAd(index);
			return (String) ad[0];
		}

		private Object getAdIcon(int index) {
			Object[] ad = getAd(index);
			if(ad[2]==null) {
				return "icon.png";
			} else {
				return ad[2];
			}
		}

		private URL getAdURL(int index) {
			Object[] ad = getAd(index);
			return (URL) ad[1];
		}
		
		private String restrictWidth(String title, String fontString, int width) {
			if(width <= 10 || title == null) return title;
			FontInfo fontInfo = ((WebBrowserApp)getApp()).getFontInfos().get(fontString);
			if(fontInfo != null) {
				while(fontInfo.measureTextWidth(title)>width) {
					title = title.substring(0, title.length()-1);
				}
				return title;
			} else {
				int pxPerChar = 12;
				if(title.length() > width/pxPerChar) {
					return title.substring(0, width/pxPerChar);
				}
			}
			return title;
		}

		@Override
		public boolean handleKeyPress(int code, long rawcode) {
			switch (code) {
			case KEY_ADVANCE:
				if (size() <= 0)
					return false;
				if (getFocus() >= size() - 1) {
					setFocus(0, true);
				} else {
					setFocus(size() - 1, true);
				}
				getBApp().play(V49BApplication.SOUND_SELECT);
				return true;
			case KEY_INFO:
				screen.showWebPageMenu();
				return true;
			case KEY_SELECT:
			case KEY_FORWARD:
				postEvent(new BEvent.Action(this, "LINK")); //$NON-NLS-1$
				return true;
			case KEY_ENTER:
				postEvent(new BEvent.Action(this, "edit")); //$NON-NLS-1$
				return true;
			case V49IHmeProtocol.KEY_OPT_ACTION_B:
			case KEY_REVERSE:
				postEvent(new BEvent.Action(this, "back")); //$NON-NLS-1$
				return true;
			case KEY_PLAY:
				postEvent(new BEvent.Action(this, "top")); //$NON-NLS-1$
				return true;
			case KEY_THUMBSDOWN:
				postEvent(new BEvent.Action(this, "downgrade")); //$NON-NLS-1$
				return true;
			case KEY_THUMBSUP:
				postEvent(new BEvent.Action(this, "upgrade")); //$NON-NLS-1$
				return true;
			}
			return super.handleKeyPress(code, rawcode);
		}
	}
	public BookmarkScreen(BApplication app) {
		/* BEGIN HME 1.4.1e only */
		super(app, app.getReceiverResolutionInfo().getPreferredRenderingResolution());
		/* END HME 1.4.1e only */

		// this.getBelow().setResource(Color.yellow);
		WebBrowserApp wba = (WebBrowserApp)getApp();
		Rectangle logoRegion = wba.getLogoRegion();
		BView logo = new BView(this.getBelow(), logoRegion.x, logoRegion.y, logoRegion.width, logoRegion.height);
		logo.setResource(wba.getTheme().getLogo(), RSRC_IMAGE_BESTFIT|RSRC_HALIGN_LEFT|RSRC_VALIGN_BOTTOM);
		
		Rectangle titleRegion = wba.getTitleRegion();
		BView banner = new BView(this.getNormal(), titleRegion.x, titleRegion.y, titleRegion.width, titleRegion.height);
		BView title = new BView(banner, 0, 0, banner.getWidth(), 20);
//debugOut("notice"+WebBrowserApp.NOTICE+":"+WebBrowserApp.BROWSER_NAME);
		title.setResource(
				title.createText("system-"+(getRowFontSize(this)+1)+".font", Color.white,  //$NON-NLS-1$ //$NON-NLS-2$
						//FIXME this is the constructor - we have to set this after folder is defined.
						MessageFormat.format(Messages.getString("BookmarkScreen.Title"), WebBrowserApp.NOTICE, BOOKMARK_NAME, WebBrowserApp.folderToReadablePath(folder)) //$NON-NLS-1$
						), RSRC_HALIGN_CENTER
						| RSRC_VALIGN_BOTTOM );
		BView hints = new BView(banner, 0, title.getHeight(), banner.getWidth(),
				banner.getHeight() - title.getHeight());
		
		// set text color when we call addHint
		WebScreen.setLeftHints(null, null, null, null, Color.white);
//		WebScreen.addHint(hints, new int[] {KEY_THUMBSUP}, BOOKMARK_NAME+"page");
//		WebScreen.addHint(hints, new int[] {KEY_THUMBSDOWN}, "(un)"+BOOKMARK_NAME+"page");
		WebScreen.addHintH(hints, new int[] {KEY_SELECT,KEY_FORWARD},
				MessageFormat.format(Messages.getString("BookmarkScreen.BOOKMARK_HINT_FOLLOW"), BOOKMARK_NAME));
		WebScreen.addHintH(hints, new int[] {V49IHmeProtocol.KEY_OPT_ACTION_B,KEY_REVERSE},
				MessageFormat.format(Messages.getString("BookmarkScreen.BOOKMARK_HINT_BACK"), BOOKMARK_NAME));
		WebScreen.addHintH(hints, new int[] {KEY_PLAY},
				MessageFormat.format(Messages.getString("BookmarkScreen.BOOKMARK_HINT_LIST"), BOOKMARK_NAME));
		WebScreen.addHintH(hints, new int[] {KEY_ENTER},
				MessageFormat.format(Messages.getString("BookmarkScreen.BOOKMARK_HINT_EDIT"), BOOKMARK_NAME));
		WebScreen.addHintH(hints, new int[] {KEY_THUMBSUP},
				MessageFormat.format(Messages.getString("BookmarkScreen.BOOKMARK_HINT_SAVE_LAST"), BOOKMARK_NAME));
		WebScreen.addHintH(hints, new int[] {KEY_THUMBSDOWN},
				MessageFormat.format(Messages.getString("BookmarkScreen.BOOKMARK_HINT_REMOVE"), BOOKMARK_NAME));
		WebScreen.addHintH(hints, new int[] {KEY_INFO},
				MessageFormat.format(Messages.getString("BookmarkScreen.BOOKMARK_HINT_MENU"), BOOKMARK_NAME));
		WebScreen.addHintH(hints, new int[] {KEY_CLEAR},
				MessageFormat.format(Messages.getString("BookmarkScreen.BOOKMARK_HINT_QUIT"), BOOKMARK_NAME));

		Rectangle listRegion = wba.getScrollRegion();
		BView listArea = new BView(this.getNormal(), listRegion.x, listRegion.y, listRegion.width, listRegion.height);

		int buttonHeight = getBApp().getSkin().get(H_BAR).getHeight();
		list = new BookmarkList(this, listArea, 0, 0, listArea.getWidth(),
				listArea.getHeight(), buttonHeight/2);
		list.setBarAndArrows(BAR_HANG, BAR_HANG, "back", "LINK"); //$NON-NLS-1$ //$NON-NLS-2$

		setFocusDefault(list);
		setupList();
	}

	private Entry<String, URL[]> getBookmark(int index) {
		int i = 0;
		for (Entry<String, URL[]> entry : this.bookmarks.entrySet()) {
			if (i == index)
				return entry;
			++i;
		}
		return null;
	}
	
	private void renameBookmark(int index, String title) {
		Entry<String, URL[]> entry = removeBookmark(index);
		this.bookmarks.put(title, entry.getValue());
	}
	
	private void addBookmark(URL[] url, String title) {
		this.bookmarks.put(title, url);
	}
	
	private Entry<String, URL[]> removeBookmark(int index) {
		Entry<String, URL[]> result = null;
		int i = 0;
		for (Entry<String, URL[]> entry : this.bookmarks.entrySet()) {
			if (i == index) {
				result = entry;
				break;
			}
			++i;
		}
		// can't remove folders
		if (result != null
				&& result.getValue()[0].toString().startsWith(
						WebBrowserApp.BOOKMARK_FOLDER_PREFIX)) {
			return null;
		}
		this.bookmarks.remove(result.getKey());
		return result;
	}
	
	private String lastVisitedTitle = ""; //$NON-NLS-1$
	private URL lastVisitedURL = null;
	public boolean isFirstPage;
	public String folder;
	private String media;
	
//	/** Deal with a bug in BList focus handling. */
//	@Override
//	public void setFocus(BView focus) {
//		try {
//			super.setFocus(focus);
//		} catch(NullPointerException ex) {
//			debugOut("NPE during setFocus of "+focus);
//		}
//	}

	@Override
	public boolean handleEnter(Object arg, boolean isReturn) {
		((WebBrowserApp)getBApp()).hidePleaseWait();
		showBackground();
		
		if(isReturn) {
			if(handleReturnArg(arg))
				return true;
			// else
			// continue with re-display if this is just returning from a web
			// page (only happens when this is the very first screen)
		}
		if(arg instanceof String) {
			String folder = ((String)arg);//((WebBrowserApp) getBApp()).getFolder(((String)arg));
debugOut("bookmark folder:"+folder);
//			if(this.folder != null) {
//				this.folder += WebBrowserApp.BOOKMARK_FOLDER_SEPARATOR+folder;
//			} else {
				this.folder = folder;
//			}
		}
		if(this.folder == null) {
			this.folder = "";
		}
		this.bookmarks = ((WebBrowserApp) getBApp()).getBookmarks(this.folder);
		this.lastVisitedURL = ((WebBrowserApp) getBApp()).getLastVisited();
		this.lastVisitedTitle = ((WebBrowserApp) getBApp()).getLastVisitedTitle();
		if(lastVisitedURL == null) {
			try {
				lastVisitedURL = new URL("http://www.google.com"); //$NON-NLS-1$
				lastVisitedTitle = "(sample)Google"; //$NON-NLS-1$
			} catch (MalformedURLException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		setupList();

		return super.handleEnter(arg, isReturn);
	}
	
	private void showWebPageMenu() {
		this.media = ((WebBrowserApp)getBApp()).getStreamSourceCustomURL();
		BScreen menu = new WebMenuScreen(getBApp(), this.media);

		// pass the bookmark title in so the menu can offer to let you edit it.
		Object arg = null;
		int index = list.getFocus();
		if (index < this.bookmarkStartIndex) {
		} else {
			index -= this.bookmarkStartIndex;
			Entry<String, URL[]> bookmark = getBookmark(index);
			if (bookmark != null
					&& !bookmark.getValue()[0].toString().startsWith(WebBrowserApp.BOOKMARK_FOLDER_PREFIX)) {
				arg = bookmark.getKey();
			}
		}
		
		// can't use TRANSITION_FORWARD because many menu commands cause a
		// screen replacement which fires during handleEnter BEFORE the pop
		// transition is processed
		getBApp().push(menu, TRANSITION_NONE, arg);
	}
	
	public void showBackground() {
//		if(((WebBrowserApp)getBApp()).isStreamTypeVideo()) {
			Object resource = new Color(80, 0, 80);
			try {
				resource = getBApp().getSkin().get("showcases").getResource();
			} catch (Exception e) {}
			getBelow().setResource(resource, RSRC_IMAGE_BESTFIT);
//		} else {
//			StreamResource videoLoop = createStream(WebBrowserApp.STREAM_LOOP_PURPLE_SHOWCASES, "video/mpeg", true); //$NON-NLS-1$
//			getBelow().setResource(videoLoop);
//			flush();
//		}
	}
	public void disposeBackground() {
		if(getBelow().getResource() != null && getBelow().getResource() instanceof StreamResource
				&& ((StreamResource)getBelow().getResource()).getURI().equals(WebBrowserApp.STREAM_LOOP_PURPLE_SHOWCASES)) {
			getBelow().getResource().remove();
			getBelow().clearResource();
			flush();
		}
	}

	@Override
	public boolean handleExit() {
		disposeBackground();
		
		// Note, if we push during a pop this still happens,
		// so be sure to turn off isPopping temporarily if we do a push
		if(((WebBrowserApp)getBApp()).isPopping) {
			// clean up everything we created on entering
			removeAllChildren(this.list);
			this.list.remove();
			// clean up some automatic stuff (especially the focus views)
			//TODO remove actual above/normal/below views themselves?
			//TODO just bite the bullet and removeAllChildren(screen)?
			//TODO make sure this is doing the right thing for resource viewing cases.
			removeAllChildren(getAbove());
			removeAllChildren(getNormal());
			removeAllChildren(getBelow());
			
			return true;
		}
		return super.handleExit();
	}
	
	private void removeAllChildren(View parent) {
		// collect the views before removing them, otherwise the getChild call
		// gets thrown by the remove calls.
		ArrayList<View> children = new ArrayList<View>();
		for(int i = 0 ; i < parent.getChildCount() ; ++i) {
			children.add( parent.getChild(i));
		}

		for(View child : children) {
			removeAllChildren(child);
			if(child.getResource() != null) {
				if(child.getResource().getClass().getName().contains("Text")) { //$NON-NLS-1$
					child.getResource().remove();
				}
			}
			child.clearResource(); // resource's own finalize will remove itself from tivo
			// make sure focus gets cleared if the one we're removing has it.
			if(child.hasFocus()) {
				this.setFocus(null);
			}
			child.remove();
		}
	}

	private static final String ENTER_URL = Messages.getString("BookmarkScreen.Enter_URL"); //$NON-NLS-1$
	private static final String ADVERTISEMENT = "\nyour ad here\n"; //$NON-NLS-1$
	public static final String BOOKMARK_NAME = Messages.getString("BookmarkScreen.BOOKMARK_NAME"); //$NON-NLS-1$

	private void setupList() {
		int oldFocus = list.getFocus();
		if(oldFocus < 0) oldFocus = 0;
		list.clear();
		// add other entries and move bookmarkstartindex
		if(lastVisitedURL != null) {
			list.add(MessageFormat.format(Messages.getString("BookmarkScreen.Last_Visited"),lastVisitedTitle)); //$NON-NLS-1$
		}
		list.add(ADVERTISEMENT);
		list.add(ENTER_URL);
		this.bookmarkStartIndex = 3;//list.size();
		// make the BNumberedList start numbering at the bookmarks
		list.setPositionOffset(this.bookmarkStartIndex);

		if (bookmarks != null) {
			list.add(new Vector<String>(bookmarks.keySet()));
		}
		if(oldFocus >= list.size()) {
			oldFocus = list.size()-1;
		}
		list.setFocus(oldFocus, false);
	}

	@Override
	public boolean handleAction(BView view, Object action) {
		if (action.equals("LINK")) { //$NON-NLS-1$
			int index = list.getFocus();
			if (index < this.bookmarkStartIndex) {
				if(list.get(index).equals(ADVERTISEMENT)) {
					URL url = list.getAdURL(index);
					if (url == null) {
						return true;
					} else {
						((WebBrowserApp)getBApp()).logBookmark("thumztakVisitAd", url.toString(), null, folder, null);
						// could be an ad for another actual app!
						if (url.getProtocol().equals("hme")) { //$NON-NLS-1$
							((WebBrowserApp) getBApp()).loadApp(url);
						} else {
							replaceBookmarkScreen(url);
						}
					}
				} else if(list.get(index).equals(ENTER_URL)) {
					setUpKeyboard();
					
					if(isTypo) {
						typo.setKeyboard(Keyboard.URL);
						typo.setActions("newURL", "urlCancel"); //$NON-NLS-1$ //$NON-NLS-2$
						typo.setValue(this.typedURL);
					} else {
						bkey.setValue(this.typedURL);
						cancel = new ActionButton(getAbove(), ((WebBrowserApp)getApp()).getSafeTitleHorizontal(), 150, 200, 20, "urlCancel"); //$NON-NLS-1$
						cancel.setBarAndArrows("urlCancel", "right", null, "down", false); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
						new BView(cancel, 5, 5, cancel.getWidth()-10, cancel.getHeight()-10).setResource(cancel.createText("default-"+getRowFontSize(this)+".font", Color.white, "Cancel")); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
						ok = new ActionButton(getAbove(), ((WebBrowserApp)getApp()).getSafeTitleHorizontal()+250, 150, 200, 20, "newURL"); //$NON-NLS-1$
						ok.setBarAndArrows("left",  "newURL", null, "down", false); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
						new BView(ok, 5, 5, ok.getWidth()-10, ok.getHeight()-10).setResource(ok.createText("default-"+getRowFontSize(this)+".font", Color.white, "Ok")); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					}
				} else {
					// last visited.
					
					replaceBookmarkScreen(lastVisitedURL);
//					// replace the bookmark screen with the web screen
//					// if we're the first screen, pop does nothing.
////					getBApp().pop();
////					getBApp().push(
//					if(isFirstPage) {
//						getBApp().push(
//								new WebScreen(getBApp()), TRANSITION_FORWARD,
//								lastVisitedURL);
//					} else {
//					getBApp().replaceCurrentScreen(
//							new WebScreen(getBApp()), TRANSITION_FORWARD,
//							lastVisitedURL);
//					}
					
				}
			} else {
				index -= this.bookmarkStartIndex;
				Entry<String, URL[]> bookmark = getBookmark(index);
				if(bookmark.getValue()[0].getProtocol().equals("hme")) { //$NON-NLS-1$
					((WebBrowserApp)getBApp()).loadApp(bookmark.getValue()[0]);
					return true;
				} else if(bookmark.getValue()[0].toString().startsWith(WebBrowserApp.BOOKMARK_FOLDER_PREFIX)) {
					String folder = bookmark.getValue()[0].toString();
					getBApp().push(new BookmarkScreen(getBApp()), TRANSITION_FORWARD, folder);
					return true;
				}
				((WebBrowserApp)getBApp()).logBookmark("thumztakVisit", bookmark.getValue()[0].toString(), bookmark.getValue()[1]==null?null:bookmark.getValue()[1].toString(), folder, bookmark.getKey());
				replaceBookmarkScreen(bookmark.getValue());
			}
			return true;
		} else if (action.equals("top")) { //$NON-NLS-1%
			// if this is top folder already, do same as "back"
			// otherwise, go to top folder.
			if (this.folder == null || this.folder.equals("") || this.folder.equals("/")) {
				if (isFirstPage) {
					((WebBrowserApp) getBApp()).logAppEnd("top");
					((WebBrowserApp) getBApp()).closeBrowser();
				} else {
					getBApp().pop();
				}
			} else {
				//TODO if we're in a folder hierarchy and top is in it, just do "back" until we're there.
				String folder = "";
				getBApp().push(new BookmarkScreen(getBApp()), TRANSITION_FORWARD, folder);
			}
			return true;
		} else if (action.equals("back")) { //$NON-NLS-1$
			if(isFirstPage) {
				((WebBrowserApp) getBApp()).logAppEnd("back");
				((WebBrowserApp)getBApp()).closeBrowser();
			} else {
				if(getBApp().getCurrentScreen() == null) debugOut("current screen is null! BEFORE pop!"); //$NON-NLS-1$
				getBApp().pop();
				if(getBApp().getCurrentScreen() == null) debugOut("current screen is null! after pop"); //$NON-NLS-1$
			}
			return true;
		} else if (action.equals("edit")) { //$NON-NLS-1$
			editBookmark();
			return true;
		} else if (action.equals("downgrade")) { //$NON-NLS-1$
			int index = list.getFocus();
			if (index < this.bookmarkStartIndex) {
			} else {
				//TODO undo further upgrades when added
				index -= this.bookmarkStartIndex;
				// remove the bookmark from the list
				Entry<String, URL[]> bookmark = removeBookmark(index);
				if (bookmark != null) {
					// remove the bookmark from storage
					((WebBrowserApp) getBApp()).logBookmark("thumztakdelete", bookmark.getValue()[0]==null?null:bookmark.getValue()[0].toString(), bookmark.getValue()[1]==null?null:bookmark.getValue()[1].toString(), folder, bookmark.getKey());
					((WebBrowserApp) getBApp()).removeBookmark(
							bookmark.getValue()[0], bookmark.getKey(), bookmark.getValue()[1]);
					// update the list view
					// this clears the list and rebuilds it, attempting to maintain focus index.
					setupList();
					return true;
				} else {
					getBApp().play(V49BApplication.SOUND_BONK);
				}
			}
			return false;

		} else if (action.equals("upgrade")) { //$NON-NLS-1$
			int index = list.getFocus();
			if (index < this.bookmarkStartIndex) {
				if(index == 0) {
					String title = this.lastVisitedTitle;
					if(title.indexOf(':')>=0) {
						title = title.substring(title.indexOf(':')+1).trim();
					}
					((WebBrowserApp) getBApp()).saveBookmark(this.lastVisitedURL, title, this.folder, null); // FIXME icons
					addBookmark(new URL[]{this.lastVisitedURL, null}, title);
					// update the list view
					// this clears the list and rebuilds it, attempting to maintain focus index.
					setupList();
					list.setFocus(list.size()-1, true);
					return true;
				} else {
					String text = String.valueOf(list.get(index));
					if(text.equals(ADVERTISEMENT)) {
						String title = list.getAdTitle(index);
						URL url = list.getAdURL(index);
						URL icon = null;
						if(list.getAdIcon(index) instanceof URL) {
							icon = (URL) list.getAdIcon(index);
						}
						// has to have a title.
						if(title != null && title.trim().length() > 0) {
							((WebBrowserApp) getBApp()).saveBookmark(url, title, this.folder, icon);
							addBookmark(new URL[]{url, icon}, title);
							// update the list view
							// this clears the list and rebuilds it, attempting to maintain focus index.
							setupList();
							list.setFocus(list.size()-1, true);
							return true;
						}
					}
					//TODO allow arbitrary url entry as bookmark
					return false;
				}
			} else {
				index -= this.bookmarkStartIndex;
				// remove the bookmark from the list
				Entry<String, URL[]> bookmark = getBookmark(index);
				if (bookmark != null) {
					String title = bookmark.getKey();
					if (title.startsWith("(temp")) { //$NON-NLS-1$
						int end = title.indexOf(')');
						if (end >= 0) {
							title = title.substring(end + 1);
						} else {
							//TODO add additional upgrade
							return false;
						}
						renameBookmark(index, title);
					} else {
						//TODO add additional upgrade
						return false;
					}
					// save the bookmark in storage
					((WebBrowserApp) getBApp()).saveBookmark(
							bookmark.getValue()[0], title, this.folder, bookmark.getValue()[1]);
					// update the list view
					// this clears the list and rebuilds it, attempting to maintain focus index.
					setupList();
					list.setFocus(list.size()-1, true);
					return true;
				}
			}
			return false;

		} else if(action.equals(BNumberedList.ACTION_LIST_ORDER_CHANGE)) {
			int originalIndex = list.getFocus();
			int originalPosition = originalIndex-list.getPositionOffset()+1;
			int newPosition = list.getRequestedPosition();
			int newIndex = newPosition-1+list.getPositionOffset();
			// index is reduced by removal of the original if we're moving down.
			if(originalPosition < newPosition)
				newIndex--;
			
			// move it, but don't rename it.
			((WebBrowserApp)getBApp()).moveBookmark(this.folder, originalPosition-1, newPosition-1, null);			
			// TODO more efficient to move item within this.bookmarks
			this.bookmarks = ((WebBrowserApp) getBApp()).getBookmarks(this.folder);
			
			// update the list view
			// this clears the list and rebuilds it, attempting to maintain focus index.
			setupList();
			list.setFocus(newIndex, true);
			return true;
		}
		return super.handleAction(view, action);
	}

	private void editBookmark() {
		int index = list.getFocus();
		if (index < this.bookmarkStartIndex) {
		} else {
			index -= this.bookmarkStartIndex;
			Entry<String, URL[]> bookmark = getBookmark(index);
			// protect against this this getting called during a pop
			boolean wasPopping = ((WebBrowserApp)getBApp()).isPopping;
			((WebBrowserApp)getBApp()).isPopping = false;
			try {
				getBApp().push(new BookmarkFolderScreen(getBApp(), this.folder, index), TRANSITION_FORWARD, bookmark);
			} finally {
				((WebBrowserApp)getBApp()).isPopping = wasPopping;
			}
//			setUpKeyboard();
//			if(isTypo ) {
////				//					this.getAbove().setResource(Color.darkGray);
////				this.getNormal().setTransparency(0.5f);
////				typo = new TyPo(this.getAbove(), getApp().getSafeTitleHorizontal(), 200, this
////						.getAbove().getWidth() - (2 * getApp().getSafeTitleHorizontal())-75, 15,
////						true);
////				typo.setExitOptions(TyPo.EXIT_CANCEL, TyPo.EXIT_FINISHED,
////						true, true, false, false);
////				typo.setActivateOnFocus(true);
////				this.setFocus(typo);
////				typo.setFocus();
////				// prevent NPE when typo's list clears and set focus to the default
////				setFocusDefault(typo);
//				
//				typo.setKeyboard(Keyboard.DEFAULT);
//				typo.setActions("editTitle", "editCancel"); //$NON-NLS-1$ //$NON-NLS-2$
//				typo.setValue(bookmark.getKey());
//			}//TODO add non-typo if needed.
		}
	}

	private boolean handleReturnArg(Object arg) {
		if(arg == null) return false;
		String action = arg.toString();
		WebBrowserApp web = (WebBrowserApp) getBApp();
		if(action.startsWith(WebMenuScreen.MENU_PREFIX)) {
			web.handleMenuAction(action);
////			} else if(action.equals("menu_brightness")) {
////				if(shade.getTransparency() == 1.0f) {
////					shade.setTransparency(0.75f);
////				} else {
////					shade.setTransparency(1.0f);
////				}
//			} else 
			if(action.equals(WebMenuScreen.MENU_PIP)) {
				disposeBackground();
				web.setStartupStreamLiveTV();
				showBackground();
			} else if(action.equals(WebMenuScreen.MENU_NO_PIP)) {
				disposeBackground();
				web.clearStartupStream();
				showBackground();
			} else if(action.equals(WebMenuScreen.MENU_PIP_MEDIA)) {
				disposeBackground();
				web.setStartupStreamCustom(this.media);
				showBackground();
////			} else if(action.equals("menu_link_thumbtack")) {
////				bookmarkFocusedLink();
			} else if(action.equals(WebMenuScreen.MENU_EDIT_LINK)) {
				editBookmark();
			}
		}
		else
	if (action.equals("newURL")) { //$NON-NLS-1$
		if(isTypo) {
			if(typo != null)
				this.typedURL = typo.getValue();
		} else {
		this.typedURL = bkey.getValue();
		}
		URL url = null;
		try {
			url = new URL(typedURL);
		} catch (IllegalArgumentException e) {
			getBApp().play(V49BApplication.SOUND_BONK);
		} catch (MalformedURLException e) {
			getBApp().play(V49BApplication.SOUND_BONK);
		}

		if (url != null) {
			((WebBrowserApp)getBApp()).logBookmark("manualURL", url.toString(), null, null, null);
			replaceBookmarkScreen(url);
//			// replace the bookmark screen with the web screen
//			// if we're the first screen, pop does nothing.
////			getBApp().pop();
////			getBApp().push(
//			if(isFirstPage) {
//				getBApp().push(
//						new WebScreen(getBApp()), TRANSITION_FORWARD,
//						url);
//			} else {
//			getBApp().replaceCurrentScreen(
//					new WebScreen(getBApp()), TRANSITION_FORWARD,
//					url);
//			}
		}
		shutDownKeyboard();
	} else if (action.equals("urlCancel") ) { //$NON-NLS-1$
		shutDownKeyboard();
	}
		return true;
	}
	
	/**
	 * show please wait, prepare new screen, and push it (no longer replaces,
	 * just pushes new screen)
	 */
	private void replaceBookmarkScreen(Object urls) {
		// replace the bookmark screen with the web screen
		((WebBrowserApp)getBApp()).showPleaseWait();
		// if we're the first screen, pop does nothing.
//				getBApp().pop();
//				getBApp().push(
		WebScreen webScreen = WebScreen.newWebScreen(getBApp());
		boolean pageSuccess = webScreen.prepare(urls);
		if(!pageSuccess) {
			((WebBrowserApp)getBApp()).hidePleaseWait();
			webScreen.remove();
			return;
		}
//		if(isFirstPage) {
		// push will exit this screen.. if replacebookmarkscreen is called during a POP,
		// that will cause trouble, so hide the fact of a pop while doing this push.
		boolean wasPopping = ((WebBrowserApp)getBApp()).isPopping;
		((WebBrowserApp)getBApp()).isPopping = false;
		try {
			getBApp().push(
					webScreen, TRANSITION_FORWARD,
					urls);
		} finally {
			((WebBrowserApp)getBApp()).isPopping = wasPopping;
		}
//		} else {
//			getBApp().pop();
//			getBApp().push(
////		getBApp().replaceCurrentScreen(
//				webScreen, TRANSITION_FORWARD,
//				urls);
//		}
	}

	private void setUpKeyboard() {
		if(isTypo ) {
			//					this.getAbove().setResource(Color.darkGray);
			this.getNormal().setTransparency(0.5f);
			if(typo == null) {
				typo = ((WebBrowserApp)getBApp()).showTyPo();
//				TyPo.setUserWordListAllowed(false);
//				typo = new TyPo(this.getAbove(), ((WebBrowserApp)getApp()).getSafeTitleHorizontal(), 200, this
//						.getAbove().getWidth() - (2 * ((WebBrowserApp)getApp()).getSafeTitleHorizontal())-75, 15,
//						true);
			}
//			typo.setExitOptions(TyPo.EXIT_CANCEL, TyPo.EXIT_FINISHED,
//					true, true, false, false);
			
//			typo.setActivateOnFocus(true);
//			this.setFocus(typo);
//			typo.setFocus();
//			// prevent NPE when typo's list clears and set focus to the default
//			setFocusDefault(typo);

			//					getBApp().push(new StringGetterScreen(null, getBApp()),
			//							TRANSITION_FADE,
			//							new String[] { "Enter URL", this.typedURL });

		} else {
			this.getNormal().setVisible(false, getResource("*500")); //$NON-NLS-1$
			bkey = new BKeyboard(getAbove(), ((WebBrowserApp)getApp()).getSafeTitleHorizontal(), 200, this.getAbove().getWidth() - (2*((WebBrowserApp)getApp()).getSafeTitleHorizontal()), this.getAbove().getHeight()-200-((WebBrowserApp)getApp()).getSafeTitleVertical(), BKeyboard.PLAIN_KEYBOARD, false);
			this.setFocus(bkey);
			bkey.setFocus();
		}
	}

	private void shutDownKeyboard() {
		this.setFocus(list);
		if(isTypo) {
//			// change default back
//			setFocusDefault(list);
//				this.getAbove().setResource(null);
			this.getNormal().setTransparency(0.0f);
			((WebBrowserApp)getBApp()).closeTyPo();
//			typo.setVisible(false);
//			    typo.remove(); // now it's a app-wide resource
			typo = null;
		} else {
			this.getNormal().setVisible(true);
			bkey.remove();
			cancel.remove();
			ok.remove();
			bkey = null;
		}
	}

}
