package com.blackledge.david.html.hme;

import java.awt.Color;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;

import com.tivo.hme.bananas.BEvent;
import com.tivo.hme.bananas.BHighlights;
import com.tivo.hme.bananas.BList;
import com.tivo.hme.bananas.BScreen;
import com.tivo.hme.bananas.BSkin.Element;
import com.tivo.hme.bananas.BView;
import com.tivo.hme.sdk.Resource;
import com.tivo.hme.sdk.View;

public class FormSelect extends BView {

	public class BSelectList extends BList {

		public ArrayList<String> listDisplayValues;
//		private BScrollBar scroll;

		public BSelectList(BView parent, int x, int y, int width, int totalHeight,
				int rowHeight) {
			super(parent, x, y, width, totalHeight, rowHeight);
//			this.setBarAndArrows(BAR_DEFAULT, BAR_DEFAULT, "cancel", "select");
			// I want the arrows outside the bar instead
	        getRowHighlights().setBarAndArrows(BAR_DEFAULT, BAR_DEFAULT, "cancel", "select", false);
	        getHighlights().setPageHint(H_PAGEUP, A_RIGHT, A_TOP);
			getHighlights().setPageHint(H_PAGEDOWN, A_RIGHT, A_BOTTOM);
	        
//	        scroll = new BScrollBar(this, parent);
		}

		@Override
		protected void createRow(BView parent, int index) {
			if(listDisplayValues != null && index < listDisplayValues.size()) {
				BView bg = new BView(parent, 0, 0, this.getWidth(), this.getRowHeight());
				bg.setResource(Color.blue);
				bg.setTransparency(0.5f);
				BView row = new BView(parent, 0, 0, this.getWidth(), this.getRowHeight());
				row.setResource(row.createText("default-15.font", Color.white,
						(String) listDisplayValues.toArray()[index]), RSRC_HALIGN_LEFT | RSRC_VALIGN_CENTER);
			}
		}
		
		@Override
		public boolean handleKeyPress(int code, long rawcode) {
			switch (code) {
			case KEY_SELECT:
			case KEY_RIGHT:
				selectCommit();
				return true;
			case KEY_UP:
			case KEY_CHANNELUP:
				if(getFocus() > 0) {
					boolean result = super.handleKeyPress(code, rawcode);
//					scroll.updateScroller(displayValues.size(), getTop(), getTop()+getNVisibleRows(), getResource(ANIM));
					return result;
				} else {
					return true;
				}
			case KEY_DOWN:
			case KEY_CHANNELDOWN:
				if(getFocus() < listDisplayValues.size()-1) {
					boolean result = super.handleKeyPress(code, rawcode);
//					scroll.updateScroller(displayValues.size(), getTop(), getTop()+getNVisibleRows(), getResource(ANIM));
					return result;
				} else {
					return true;
				}
			default:
				cancelSelect();
				return true;
			}
		}
		
		@Override
		public void setVisible(boolean visible, Resource animation) {
//			scroll.setVisible(visible, animation);
			super.setVisible(visible, animation);
		}
		
		private void selectCommit() {
			int i = this.getFocus();
			//FIXME support multiselect
			if (i >= 0) {
				ArrayList<Integer> indices = new ArrayList<Integer>();
				indices.add(i);
				FormSelect.this.closeSelect(listDisplayValues.get(i), indices);
			} else
				cancelSelect();
		}

		private void cancelSelect() {
			FormSelect.this.closeSelect(null, null);
		}
		
		@Override
		public boolean handleAction(BView view, Object action) {
			if(action.equals("select")) {
				selectCommit();
			} else {
				cancelSelect();
			}
			return true;
		}
	}

	private Object form;
	private String name;
	private Object action;
	private BView label;
	private BSelectList selectList;
	private ArrayList<String> displayValues;
	String currentDisplay;
	ArrayList<Integer> selectedIndices;

	public FormSelect(Object object, String name, BView parent, int x, int y, int width, int height) {
		// +(width%2==1?1:0) :  to adjust evenness of width to prevent overlap of left/right halves
		super(parent, x, y, getWidth(width, ((WebBrowserApp)parent.getBApp()).getTheme())
				, getHeight(height, ((WebBrowserApp)parent.getBApp()).getTheme()));
		this.name = name;
		init(object);
	}

	public static int getHeight(int height, WebBrowserTheme theme) {
		return height+theme.getFormOffsets(WebBrowserTheme.FORM.SELECT)[1];
//				10;
	} 

	public static int getWidth(int width, WebBrowserTheme theme) {
		return width+theme.getFormOffsets(WebBrowserTheme.FORM.SELECT)[0]
//				20+6
//		+(width%2==1?1:0)
		;
		}
	
	public void closeSelect(String selectedDisplay, ArrayList<Integer> indices) {
		this.setTransparency(0f);
		selectList.setVisible(false);
		getScreen().setFocus(this);
		
		label.setVisible(true);
		// label will be updated with the event.		
		// communicate selection
		WebBrowserApp.debugOut("Closed Select with "+indices+":"+selectedDisplay);
		if(selectedDisplay != null) {
			this.selectedIndices = indices;
			this.currentDisplay = selectedDisplay;
			postEvent(new BEvent.Action(this, getAction()));
		}

	}

	private boolean showSelectList() {
		if (selectList == null) {
			selectList = buildSelectList();
		}
		if (selectList != null) {
			label.setVisible(false);

			setSelectLocation(selectList);
			selectList.setVisible(true);
			getScreen().setFocus( selectList );
			updateSelectList();
		} else {
			return false;
		}
		this.setTransparency(.8f);
		return true;
	}

	private void updateSelectList() {
		if(this.selectedIndices.size() == 0) {
			// may need to init the selection?
			//FIXME seems like this shouldn't happen at all now with the addition of selectedIndex
			int index = displayValues.indexOf(this.currentDisplay);
			//TODO we're adding to the official selections... technically this an unofficial selection, but I think this is correct, still.
			this.selectedIndices.add(index);
		}
		if(this.selectedIndices.size() == 0) {
			selectList.setFocus(-1, false);
		} else {
			//FIXME support multi-selects
			selectList.setFocus(this.selectedIndices.get(0), false);
		}
	}

	private void createLabel() {
		if(label == null) {
			int[] offsets = ((WebBrowserApp)getBApp()).getTheme().getFormLabelOffsets(WebBrowserTheme.FORM.SELECT);
			label = new BView(this, offsets[0], offsets[1], getWidth()+offsets[2], getHeight()+offsets[3]);
					//10, 6, getWidth()-20-6, getHeight()-10);
		}
	}

	private void init(Object form) {
		this.form = form;
		WebBrowserTheme theme = ((WebBrowserApp)getBApp()).getTheme();
		
		BView barLeft = new BView(this, 0, 0, this.getWidth()/2, this.getHeight());
		//NOTE 1 pixel gap on simulator, but otherwise 1 pixel overlap on TiVo
		BView barRight = new BView(this, this.getWidth()/2
//				/**/+1
				, 0, this.getWidth()-this.getWidth()/2, this.getHeight());
		barLeft.setResource(theme.getFormResource(WebBrowserTheme.FORM.SELECT), RSRC_HALIGN_LEFT| RSRC_VALIGN_CENTER|RSRC_IMAGE_VFIT);
		barRight.setResource(theme.getFormResource(WebBrowserTheme.FORM.SELECT), RSRC_HALIGN_RIGHT| RSRC_VALIGN_CENTER|RSRC_IMAGE_VFIT);

		int[] details = theme.getFormDetailOffsets(WebBrowserTheme.FORM.SELECT);
		BView cycler = new BView(this, details[0], details[1], this.getWidth()+details[2]/*-6*/, this.getHeight()+details[3]);
		cycler.setResource(theme.getFormDetailResource(WebBrowserTheme.FORM.SELECT), details[4]|RSRC_IMAGE_VFIT);
		
		int[] offsets = theme.getFormOffsets(WebBrowserTheme.FORM.SELECT);
		setFocusable(true);
		getHighlights().setWhisperingArrow(BHighlights.H_RIGHT,
				BHighlights.A_RIGHT+offsets[2],
				BHighlights.A_CENTER, "right");
		getHighlights().setWhisperingArrow(BHighlights.H_LEFT,
				BHighlights.A_LEFT+offsets[3],  
				BHighlights.A_CENTER, "left");
		getHighlights().setWhisperingArrow(BHighlights.H_UP,
				 BHighlights.A_CENTER, 
				 BHighlights.A_TOP+offsets[4], "up");
		getHighlights().setWhisperingArrow(BHighlights.H_DOWN,
				 BHighlights.A_CENTER, 
				 BHighlights.A_BOTTOM+offsets[5], "down");
	}

	public Object getForm() {
		return form;
	}
	
	public String getName() {
		return name;
	}
	
	@Override
	public boolean handleKeyPress(int code, long rawcode) {
		if(code == KEY_SELECT) {
			return showSelectList();
		} else {
			return super.handleKeyPress(code, rawcode);
		}
	}

	private BSelectList buildSelectList() {
//		View ancestor = this.getParent().getParent();
		View ancestor = this.getParent();
		while(ancestor != null && !(ancestor.getParent() instanceof BScreen)) {
			ancestor = ancestor.getParent();
		}
		if(ancestor == null)
			ancestor = this.getParent().getParent();

		int selectHeight = (int) (this.getHeight()*5.5);
		BSelectList sel = new BSelectList((BView) ancestor, 0, 0,
				this.getWidth(), selectHeight, this.getHeight());
		setSelectLocation(sel);
		sel.listDisplayValues = this.displayValues;
		sel.add(displayValues.toArray());
		return sel;
	}

	private void setSelectLocation(BSelectList sel) {
		View ancestor = sel.getParent();
		int selectHeight = sel.getHeight();
		Point position = getViewPositionRelativeToAncestor(this, ancestor);
		int yOffset = getSelectOffset(ancestor, position, selectHeight);
		int newY = (int)position.getY()+yOffset;
		int newX = (int)position.getX();
		if (sel.getY() != newY || sel.getX() != newX) {
			sel.setLocation(newX, newY);
		}
	}

	private int getSelectOffset(View ancestor, Point position, int selectHeight) {
		int yOffset = this.getHeight()/2;
		if(position.getY()+yOffset+selectHeight > ancestor.getHeight()) {
			yOffset = (int) (ancestor.getHeight() - selectHeight - position.getY());
		}
		return yOffset;
	}

	private Point getViewPositionRelativeToAncestor(View view, View ancestor) {
		int x = 0;
		int y = 0;
		while(view != ancestor) {
			x += view.getX();
			y += view.getY();
			view = view.getParent();
			// include parent's translation in the calculation
			if(view != null) {
				x += view.getTranslationX();
				y += view.getTranslationY();
			}
		}
		return new Point(x, y);
	}

//	@Override
//	public boolean handleKeyRepeat(int code, long rawcode) {
//		if(code == KEY_SELECT) {
//			//FIXME do a real impl so they can go backwards, too
//			// important to support repeat since there could be a lot of things to tick through with select button in current impl
//			postEvent(new BEvent.Action(this, getAction()));
//			return true;
//		} else {
//			return super.handleKeyPress(code, rawcode);
//		}
//	}
//
	private Object getAction() {
		return this.action;
	}
	
	public void setAction(Object action) {
		this.action = action;
	}
	
	@Override
	public boolean handleAction(BView view, Object action) {
		if(selectList != null && selectList.getVisible()) {
			// eat any leftovers (mainly when the BList is at top or bottom so the
			// up/down doesn't forward to the screen's other links/buttons)
			return true;
		} else {
			return super.handleAction(view, action);
		}
	}

	public void updateSelect(LinkedHashMap<String, String> valuesDisplays, ArrayList<Integer> selectedIndices, Color labelColor, String fontString) {
		createLabel();
		if(labelColor == null) labelColor = Color.white;
		setDisplayValues( valuesDisplays.values() );
		Iterator<String> valIter = valuesDisplays.values().iterator();
		int index = 0;
		String display = "";
		while(valIter.hasNext()) {
			String disp = valIter.next();
			if(index == 0)
				display = disp;
			if(selectedIndices.contains(index)) {
				display = disp;
			}
			++index;
		}
//WebBrowserApp.debugOut("updating select with "+selected+" "+display);
WebBrowserApp.debugOut("set "+this.name+"="+selectedIndices+" ~ "+display);
		this.selectedIndices = selectedIndices;
		this.currentDisplay = display;
		if (label.getResource() != null) {
			label.getResource().remove();
//			label.clearResource();
		}
		label.setResource(label.createText(fontString, labelColor, display));
	}

	private void setDisplayValues(Collection<String> values) {
		this.displayValues = new ArrayList<String>(values);
		if(selectList != null) {
			selectList.listDisplayValues = this.displayValues;
		}
	}
	
}
