package com.blackledge.david.html.hme;

import java.awt.Font;
import java.awt.font.FontRenderContext;
import java.awt.font.GlyphMetrics;
import java.awt.font.GlyphVector;
import java.awt.font.LineMetrics;
import java.awt.geom.AffineTransform;
import java.util.HashMap;
import java.util.Map;

import com.tivo.hme.sdk.HmeEvent.FontInfo;

/**
 * FontInfo based on a local Java {@link Font} object.
 * @author David
 *
 */
public class LocalFontInfo extends FontInfo {

	private Font font;

	private LocalFontInfo(Font font, int id, float ascent, float descent,
			float height, float lineGap, Map glyphInfo) {
		super(id, ascent, descent, height, lineGap, glyphInfo);
		this.font = font;
	}
	
	public Font getFont() {
		return font;
	}

	public static LocalFontInfo newFontInfo(Font font) {
		return new LocalFontInfo(font, getId(font), calcAscent(font), calcDescent(font), calcHeight(font), calcLineGap(font), createGlyphInfo(font));
	}

	private static int getId(Font font) {
		//FIXME is this ok?
		return -1;
	}

	private static String lineMetricsString = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
	private static FontRenderContext lineMetricsFRC = 
	//new FontRenderContext(AffineTransform.getScaleInstance(10000.0, 10000.0), 
	//		false, false);
	new FontRenderContext(AffineTransform.getScaleInstance(3.0, 4.0),
	        true, true);
	//new FontRenderContext(new AffineTransform(),
	//        true, true);
	//new BufferedImage(1, 1, BufferedImage.TYPE_4BYTE_ABGR).createGraphics().getFontRenderContext();
//	System.out.println("Transform!"+fontRenderContext.getTransform());

	private static float calcAscent(Font font) {
		LineMetrics line = font.getLineMetrics(lineMetricsString, lineMetricsFRC);
		return line.getAscent();
	}

	private static float calcDescent(Font font) {
		LineMetrics line = font.getLineMetrics(lineMetricsString, lineMetricsFRC);
		return line.getDescent();
	}

	private static float calcHeight(Font font) {
		LineMetrics line = font.getLineMetrics(lineMetricsString, lineMetricsFRC);
		return line.getHeight();
	}

	private static float calcLineGap(Font font) {
		LineMetrics line = font.getLineMetrics(lineMetricsString, lineMetricsFRC);
		return line.getLeading();
	}

	private static Map createGlyphInfo(Font font) {
		HashMap<Character, GlyphInfo> glyphInfo = new HashMap<Character, GlyphInfo>();
		int maxGlyph = Math.min(font.getNumGlyphs(), 128);
		for (int i = 0; i < maxGlyph; ++i) {
			Character ch = Character.valueOf((char) i);
			if (font.canDisplay(ch)) {
				GlyphMetrics glyphMetrics = font.createGlyphVector(
						lineMetricsFRC, new char[] { ch }).getGlyphMetrics(0);
				glyphInfo.put(ch, new GlyphInfo(ch, glyphMetrics.getAdvance(),
						(float) glyphMetrics.getBounds2D().getWidth()));
//System.out.println(ch+": "+glyphMetrics.getAdvance()+":"+(float) glyphMetrics.getBounds2D().getWidth());
			}
		}
		return glyphInfo;
	}

}
