package com.blackledge.david.html.hme;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class VimeoHandler {

	private static final Pattern vimeoVideoPattern = Pattern.compile(".*vimeo\\.com.*/(\\d\\d\\d\\d\\d\\d+).*");
	//http://vimeo.com/68548877
	//<object ...<param name="movie" value="http://vimeo.com/moogaloop.swf?clip_id=68548877&...
	//<iframe src="http://player.vimeo.com/video/68548877?" ...
	public static boolean isVimeoVideo(URL url) {
		if(url == null) return false;
		boolean result = vimeoVideoPattern.matcher(url.toString()).matches();
//		WebBrowserApp.debugOut("VIMEO TEST:"+url+" "+result);
		return result;
	}
	
	private static final Pattern videoIdPattern = vimeoVideoPattern;
	
	/**
	 * Supposed to get a video download URL based on the content of the Vimeo
	 * page (if it has a download link)... but in October 2013 the pages were
	 * changed so creating a download URL is pretty impossible now - always
	 * returns null now because of that.
	 * 
	 * @param vimeoURL
	 *            original request URL
	 * @param referer 
	 * @param browser
	 *            instance to use to download the page correctly.
	 * @return the video download URL or null of one isn't available.
	 */
	public static URL getVideoURL(URL vimeoURL, URL referer, WebBrowserApp browser) {
		if(vimeoURL == null) return null;
		//http://stackoverflow.com/a/13999482
		
//		var player = document.getElementsByClassName("player")[0].getAttribute("id");
//		player = eval(player.replace("player_", "clip"));
//		var time = player.config.request.timestamp;
//		var sig = player.config.request.signature;
//		var clip_id = window.location.href.substring(17);
//
//		var url = "http://player.vimeo.com/play_redirect" +
//		  "?clip_id=" + clip_id +
//		  "&sig=" + sig +
//		  "&time=" + time;

		// "timestamp":1361662094,"signature":"c15b0cd5faf645c5ea09634b0cd94d70"
		Pattern videoPageDetailPattern = Pattern.compile(
				Pattern.quote("\"timestamp\"")+"\\s*:\\s*" +
						"\"?([^\",]+)\"?" +
						"\\s*,\\s*"+
				Pattern.quote("\"signature\"")+"\\s*:\\s*" +
						"\"?([^\",]+)\"?"
				);
		Matcher idMatcher = videoIdPattern.matcher(vimeoURL.toString());
		String clip_id = null;
		if(idMatcher.find() && idMatcher.groupCount() >= 1) {
			clip_id = idMatcher.group(1);
//			try {
//				URL youTubeHME = new URL("hme://204.176.49.116:7288/youtube" 
//								+ "?" 
//								+ "jumpTo=VIDEO"
//								+ "&" 
//								+ "videoid="+v
//						);
//				browser.loadApp(youTubeHME);
//				return true;
//			} catch (MalformedURLException e) {
//				// FIXME Auto-generated catch block
//				e.printStackTrace();
//			}
		}
		if(clip_id != null) {
			String vimeoPage = browser.getRawText(vimeoURL, referer);
			if(vimeoPage != null) {
				boolean downloadable = vimeoPage.indexOf("title=\"Download\"")>=0;
				if(downloadable) {
					Matcher detailMatcher = videoPageDetailPattern.matcher(vimeoPage);
					if(detailMatcher.find() && detailMatcher.groupCount() >= 2) {
						String time = detailMatcher.group(1);
						String sig = detailMatcher.group(2);
						String downloadUrl = "http://player.vimeo.com/play_redirect" +
								"?clip_id=" + clip_id +
								"&sig=" + sig +
								"&time=" + time;
						if(browser.getCurrentResolution().getWidth() > 640) {
							downloadUrl += "&quality=hd";
						}
						try {
							return new URL(downloadUrl);
						} catch (MalformedURLException e) {
							WebBrowserApp.debugOut("VIMEO : "+e);
							return null;
						}
					}
					WebBrowserApp.debugOut("VIMEO failed: detail groupCount="+detailMatcher.groupCount()+" "+vimeoPage);
				} else {
					WebBrowserApp.debugOut("VIMEO failed: no download link "+vimeoURL);
					return null;
				}
			}
		}
		WebBrowserApp.debugOut("VIMEO failed: id groupCount="+idMatcher.groupCount()+" "+clip_id+" "+vimeoURL);
		return null;
	}
	public static String getVideoMimeType() {
		return "video/mp4";
	}
	
	private static final Pattern vimeoThumbnailSmallPattern = Pattern
			.compile(".*<" + "thumbnail_small" + ">(.*)</" + "thumbnail_small" + ">.*",
					Pattern.CASE_INSENSITIVE | Pattern.DOTALL);
	private static final Pattern vimeoThumbnailLargePattern = Pattern
			.compile(".*<" + "thumbnail_large" + ">(.*)</" + "thumbnail_large" + ">.*",
					Pattern.CASE_INSENSITIVE | Pattern.DOTALL);
	/**
	 * get the URL of an image representing the video. Downloads the Vimeo API
	 * link that provides thumbnail links.
	 */
	public static URL getPoster(String vimeoSrc) {
		return getPosterSmall(vimeoSrc);
	}
	public static URL getPosterSmall(String vimeoSrc) {
		return getPoster(vimeoSrc, vimeoThumbnailSmallPattern);
	}
	public static URL getPosterLarge(String vimeoSrc) {
		return getPoster(vimeoSrc, vimeoThumbnailLargePattern);
	}
	private static URL getPoster(String vimeoSrc, Pattern vimeoThumbnailPattern) {
		Matcher idMatcher = videoIdPattern.matcher(vimeoSrc);
		String clip_id = null;
		if(idMatcher.find() && idMatcher.groupCount() >= 1) {
			clip_id = idMatcher.group(1);
			String request = "http://vimeo.com/api/v2/video/" + clip_id
					+ ".xml";

			String text;
			URL req;
			try {
				req = new URL(request);
				// usually we use the browser reference to get the page
				// contents, but that's not reasonable how this is meant to be
				// used.  Just do a basic download.
				text = getRawText(req);
				if(text == null) {
					return null;
				}
			} catch (Exception e) {
				return null;
			}

			Matcher thumbnail = vimeoThumbnailPattern.matcher(text);
			if (thumbnail.matches()) {
				String url = thumbnail.group(1);
				if (url != null && url.trim().length() > 0) {
					try {
						
						return new URL(req, url.trim());
						
					} catch (MalformedURLException e) {
						return null;
					}
				}
			}
		}
		
		return null;
	}
	
	/** get raw text without a browser */
	private static String getRawText(URL urlBase) {
		try {
			URLConnection connection = urlBase.openConnection();
			
			InputStream stream = connection.getInputStream();
			BufferedReader reader = new BufferedReader(new InputStreamReader(
					stream));
			StringBuffer result = new StringBuffer();
			for (String line = reader.readLine(); line != null; line = reader
					.readLine()) {
				result.append(line).append("\n"); //$NON-NLS-1$
			}
			stream.close();
			return result.toString();
		} catch (IOException e) {
			WebBrowserApp.debugOut("getRawText("+urlBase+"): "+e);
			return null;
		}
	}


}
