package com.blackledge.david.html.hme;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.List;
import java.util.Map;

import com.blackledge.david.tivo.V49BApplicationFactory;

import com.tivo.hme.interfaces.IApplication;
import com.tivo.hme.interfaces.IArgumentList;
import com.tivo.hme.interfaces.IContext;
import com.tivo.hme.interfaces.IHmeConstants;
import com.tivo.hme.interfaces.IHttpRequest;
import com.tivo.hme.sdk.io.FastInputStream;

public class WebBrowserAppFactory extends V49BApplicationFactory {
	private String startTitle;
	private String startURL;
	private int count;
	private static int urlCount = 0;

	/**
	 * Override to set startURL if server instantiated with a startURL argument.
	 */
	@Override
	public IApplication createApplication(IContext context)
			throws IOException {
		IApplication application = super.createApplication(context);
		
		WebBrowserApp app = (WebBrowserApp) application;
		//TODO currently this only acts as default if no URL arg, should this override instead?
		if(app.startURL == null) {
			app.startURL = this.startURL;
		}

		return application;
	}

	@Override
	public InputStream getStream(String uri) throws IOException {
        synchronized (lock) {
        	if (WebProxy.isStreamProxyUri( uri )) {
        		WebProxy proxy = getProxy(uri);
        		String urip = proxy.getProxyUri();
        		URL url = WebBrowserApp.getURL(getAssetURI(), urip);
        		URLConnection conn = url.openConnection();
        		// set headers and referer on conn
        		if(proxy.headers != null) {
        		}
        		if(proxy.referer != null) {
        			conn.addRequestProperty("Referer", proxy.referer.toString());
        		}
        		URLStream urlStream = 
        				new URLStream(
        						conn.getInputStream(), conn.getContentLength());
        		return urlStream;
            }
        	return super.getStream(uri);
        }
	}
	//FIXME set up for proxying images for apps
	
	public WebProxy createProxy(URL original, URL referer, Map<String, List<String>> headers, String contentType, boolean forceProxy) {
		WebProxy proxy = new WebProxy(original, referer, headers, contentType, forceProxy);
		//FIXME save proxy to a list, and return existing proxy for duplicate requests
		return proxy;
	}
	
	private WebProxy getProxy(String streamUri) {
		//FIXME search through list for matching Uri and return it or null
		return null;
	}
	
	// local copy of sdk's package-private class
    /**
     * A simple stream that responds correctly to in.available(). This is for
     * http streams.
     */
    static class URLStream extends FastInputStream
    {
        long contentLength;
        URLStream(InputStream in, long contentLength)
        {
            super(in, IHmeConstants.TCP_BUFFER_SIZE);
            this.contentLength = contentLength;
        }
        public int available()
        {
            return (int) contentLength;
        }

        /**
         * This is a public entry point and must be synchronized.
         *
         * @see com.tivo.hme.interfaces.IHmeThreadSafe
         */
        public int read() throws IOException
        {
            synchronized (lock) {       // inherited from superclass
                contentLength -= 1;
                return in.read();
            }
        }

        /**
         * This is a public entry point and must be synchronized.
         *
         * @see com.tivo.hme.interfaces.IHmeThreadSafe
         */
        public int read(byte b[], int off, int length) throws IOException
        {
            synchronized (lock) {
                int n = super.read(b, off, length);
                if (n > 0) {
                    contentLength -= n;
                }
                return n;
            }
        }

        /**
         * This is a public entry point and must be synchronized.
         *
         * @see com.tivo.hme.interfaces.IHmeThreadSafe
         */
        public long skip(long n) throws IOException
        {
            synchronized (lock) {
                n = super.skip(n);
                if (n > 0) {
                    contentLength -= n;
                }
                return n;
            }
        }
    } // class URLStream
	
	
	
	@Override
	protected void init(IArgumentList args) {
		synchronized (WebBrowserAppFactory.class) {
			count = urlCount;
			++urlCount;
		}
		
		if(args.getRemainingCount() > 0) {
			startURL = args.shift();
		}
		if(args.getRemainingCount() > 0) {
			// serve app with (argument) title in this case.
			startTitle = args.shift();
		}
//			super.init(args);
	}
	
	@Override
	public String getAppTitle() {
		if(this.startTitle != null) {
			return this.startTitle;
		}
		return super.getAppTitle();
	}
	
	@Override
	public String getAppName() {
		String appName = super.getAppName();
		if(count == 0) {
			return appName;
		} else {
			return appName.substring(0, appName.length()-1)+count+"/";
		}
	}
	
	@Override
	protected void addHeaders(IHttpRequest http, String uri)
			throws IOException {
		// TODO Auto-generated method stub
		super.addHeaders(http, uri);
	}
	
	@Override
	protected void close() {
		// TODO Auto-generated method stub
		super.close();
	}
}