package com.blackledge.david.html.hme.host;

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;

import com.blackledge.david.tivo.V49BApplication;

import com.tivo.hme.host.http.server.HttpRequest;
import com.tivo.hme.host.http.server.HttpServer;
import com.tivo.hme.host.sample.Listener;
import com.tivo.hme.host.util.Config;
import com.tivo.hme.interfaces.ILogger;

public class TiVoConnect extends HttpServer {

	protected final boolean DEBUG;
	protected final ILogger logger;
	protected final Listener applistener;
	public static final List<String> genres = Arrays.asList(new String[] {
			HMEItem.GENRE_VIDEO, HMEItem.GENRE_MUSIC, HMEItem.GENRE_PHOTOS,
			HMEItem.GENRE_OTHER });

	public TiVoConnect(Config config, Listener applistener, ILogger logger) throws IOException {
		super(config);
//System.out.println("TiVoConnect set up!");
        this.logger = logger;
        this.applistener = applistener;

        DEBUG = config.getBool("listener.debug", false);
//        tm0 = System.currentTimeMillis();
//        
//        config.put("http.acceptor.name", ACCEPTOR_NAME);

        start();
	}

	protected String getServerTitle() {
		return "TiVo Server";
	}

	@Override
	public void handle(HttpRequest request) throws IOException {
		if(request.getURI().contains("TiVoConnect")) {
			long start = new Date().getTime();
			System.out.println(request.getURI());
//			System.out.println(request.getInetAddress());
			System.out.println(request.getHeaders());
				//			String query = "";
				//			Iterator keys = request.getHeaders().getKeys();
				//			while (keys.hasNext()) {
				//				String key = (String) keys.next();
				//				String value = request.getHeaders().get(key);
				//				query += URLEncoder.encode(key, "UTF-8");
				//				if(value != null) {
				//					query += "="+URLEncoder.encode(value, "UTF-8");
				//				}
				//				if(keys.hasNext())
				//					query += "&";
				//			}
						
			String response;
			try {
				response = this.toString(request);
			} catch(Throwable e) {
				e.printStackTrace();
				// really not much point to this... Specific TiVo requesting is
				// probably the cause, but TiVos won't interpret this response.
				request.reply(HTTP_STATUS_INTERNAL_ERROR, e.toString());
				return;
			}
			
//System.out.println("RESPONDING with:"+response);

			request.reply(HTTP_STATUS_OK, "");
			// originally sending:
//			{Content-length=6627, 
//			Server=HttpServer}			
			// Premiere says:
//			{Transfer-Encoding=chunked, 
//			Date=Thu, 03 Jan 2013 20:49:38 GMT, 
//			Keep-Alive=timeout=5, max=100, 
//			Content-Type=text/xml, 
//			Connection=Keep-Alive, 
//			Server=Apache}
			
			request.addHeader("Content-Type", "text/xml");
			
			// Add expiration header to prevent excess caching of
			// this file.
			// Expires: Wed, 1 Jun 2011 20:00:00 GMT
			// w3.org says Thu, 01 Dec 1994 16:00:00 GMT
			SimpleDateFormat df = 
					new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz"); // proper format, I believe.
			df.setTimeZone(TimeZone.getTimeZone("GMT"));
			Date expdate = new Date();
			expdate.setTime(expdate.getTime() + getExpirationLifespan());
			request.addHeader("Expires", df.format(expdate));

			// send the actual file
			byte[] bytes = response.getBytes();
			OutputStream outputStream = request.getOutputStream(bytes.length);
			outputStream.write(bytes);
			
			long time = new Date().getTime()-start;
			if(time > 10*1000) {
				// this is not a debug line!
				System.out.println(" WARNING: TiVoConnect response took "+time+" (more than 10s) and may have timed out!");
			}
//			System.out.println("RESPONSE WRITTEN millis:"+(new Date().getTime()-start));
		} else if(!handleOtherRequest(request)){
			request.reply(HTTP_STATUS_NOT_FOUND, "Only TiVoConnect supported");
		}
							
	}

	/**
	 * Defaults to 15 minutes (5*60*1000 millis).
	 */
	protected int getExpirationLifespan() {
		return 15 * 60 * 1000;
	}

	protected boolean handleOtherRequest(HttpRequest request)
			throws IOException {
		return false;
	}

	private String toString(HttpRequest request) {
		if (request == null
				|| request.get(V49BApplication.HTTP_HEADER_TIVO_ID) == null) {
			return toString();
		}

		HMEItem[] items = getHMEItems(request);
		logger.log(ILogger.LOG_INFO, request.get(V49BApplication.HTTP_HEADER_TIVO_ID)+" items: "+items.length);
		return toString(items);
	}

	public String toString() {
			
		HMEItem[] items = getHMEItems(null);
		String response = toString(items);
		return response;
	}

	/**
	 * Override to list all HMEItems that should be listed in this TiVoConnect
	 * request. Warning, any HMEItem whos URL isn't actually running appears to
	 * render the entire listing invalid to the TiVo. You just get an IP number
	 * on the TiVo
	 * 
	 * @param request
	 * @return
	 */
	protected HMEItem[] getHMEItems(HttpRequest request) {
		return null;
	}

	private String toString(HMEItem[] items) {
		int totalItems = items == null? 0 : items.length;
//							String url = base+"";//FIXME build url
//				//			if(query != null && query.length() > 0) {
//				//				url += "?"+query;
//				//			}
//				System.out.println(url);
			
			
		String response = 
				"<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>" + "\n" +
				"  <TiVoContainer>" + "\n" +
				"   <Details>" + "\n" +
				"    <ContentType>x-container/tivo-server</ContentType>" + "\n" +
				"    <SourceFormat>x-container/folder</SourceFormat>" + "\n" +
				"    <Title>" + HMEItem.xmlEncode(getServerTitle()) + "</Title>" + "\n" +
				"    <TotalItems>" + totalItems + "</TotalItems>" + "\n" +
				"   </Details>" + "\n" +
				"   <ItemCount>" + totalItems + "</ItemCount>" + "\n" +
				"   <ItemStart>0</ItemStart>" + "\n";
		if(items != null)
		for(HMEItem item : items) {
			response += "" + item.toString();
		}
		response +=
				"  </TiVoContainer>";
		return response;
	}
	
	protected String urlEncode(String arg) {
		try {
			return URLEncoder.encode(arg, "UTF-8");
		} catch (UnsupportedEncodingException e) {
			return arg;
		}
	}

//	@Override
//	protected void handleException(Object context, Throwable t) {
//		// TODO Auto-generated method stub
//		super.handleException(context, t);
//	}

}


//			response = "\r\n" + 
//					"<TiVoContainer>\r\n" + 
//					"<Details>\r\n" + 
//					"  <ContentType>x-container/tivo-server</ContentType>\r\n" + 
//					"  <SourceFormat>x-container/folder</SourceFormat>\r\n" + 
//					"  <TotalItems>1</TotalItems>\r\n" + 
//					"  <Title>MP3tunes HME Server</Title>\r\n" + 
//					"</Details>\r\n" + 
//					"<ItemStart>0</ItemStart>\r\n" + 
//					"<ItemCount>1</ItemCount>\r\n" + 
//					"<Item>\r\n" + 
//					"  <Details>\r\n" + 
//					"    <ContentType>application/x-hme</ContentType>\r\n" + 
//					"    <SourceFormat>x-container/folder</SourceFormat>\r\n" + 
//					"    <Title>MP3tunes Locker</Title>\r\n" + 
//					"  </Details>\r\n" + 
//					"  <Links>\r\n" + 
//					"    <Content>\r\n" + 
//					"      <Url>http://168.143.122.40:7288/mp3tunes/</Url>\r\n" + 
//					"    </Content>\r\n" + 
//					"  </Links>\r\n" + 
//					"</Item>\r\n" + 
//					"</TiVoContainer>";

//			totalItems=1;
//			response = "  <TiVoContainer> \r\n" + 
//					"   <Details> \r\n" + 
//					"    <ContentType>\r\n" + 
//					"     x-container/tivo-server\r\n" + 
//					"    </ContentType> \r\n" + 
//					"    <SourceFormat>\r\n" + 
//					"     x-container/folder\r\n" + 
//					"    </SourceFormat> \r\n" + 
//					"    <Title>TiVo Server</Title> \r\n" + 
//					"    <TotalItems>\r\n" + 
//					"     " +
//					totalItems +
//					"\r\n" + 
//					"    </TotalItems> \r\n" + 
//					"   </Details> \r\n" + 
//					"   <ItemCount>\r\n" + 
//					"    " +
//					totalItems +
//					"\r\n" + 
//					"   </ItemCount> \r\n" + 
//					"   <ItemStart>\r\n" + 
//					"    0\r\n" + 
//					"   </ItemStart> \r\n" + 
		
//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Credit Card Update (PCI enabled)</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      984BBCD1-F705-6F3C-8E1A-BA7F4C26B570\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      15\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.59:7287/CreditCard_20080912-20080912_131340/1221250421/1289849840-1605209840/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/CreditCard_20080912-20080912_131340/1221250421/1289849840-1605209840/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Buy a TiVo Premiere</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      0F64DC01-096D-D7A9-1145-060E26DD0CE5\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      3\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.59:7287/TivoUpgrade-20120424_122953/1335295796/1335399021-1650759021/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/TivoUpgrade-20120424_122953/1335295796/1335399021-1650759021/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" +

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Amazon v1.9 (2010.10.25-0632)</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      35FE011C-3850-2228-FBC5-1B9EDBBE5863\r\n" + 
//					"     </Uuid> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/amazon-1-9.2011.03.16-0632-20110316_154048/1300315250/1317939394-1633299394/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/amazon-1-9.2011.03.16-0632-20110316_154048/1300315250/1317939394-1633299394/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Order Domino's Pizza Now</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      6BA2A86B-619B-67A6-BFA3-0E137B167A5A\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      21\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/dominos.2009.05.07.r1021.logerror.signed-20090507_174812/1241743696/1289849802-1605209802/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/dominos.2009.05.07.r1021.logerror.signed-20090507_174812/1241743696/1289849802-1605209802/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Browse and Buy Movie Tickets</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      B9DDBECF-BC6D-5E33-388A-4CF840068D79\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      30\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/fandivo-1.1.2011.11.10-1645-20111115_112508/1321385109/1343238702-1658598702/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/fandivo-1.1.2011.11.10-1645-20111115_112508/1321385109/1343238702-1658598702/icon_SD.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Photos &amp; Slideshows</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      BBC7824A-40C6-8B61-4D87-6A3D799F365B\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      47\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/hdphotos-ml_2010.09.13-1339-20100913_134517/1284410719/1343239081-1658599081/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/hdphotos-ml_2010.09.13-1339-20100913_134517/1284410719/1343239081-1658599081/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Hme Down</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      9D99EC4B-B08C-191C-3255-40B58B1F5DAF\r\n" + 
//					"     </Uuid> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/hmedown_20080317-20080402_170823/1207181304/1240249987-1587405187/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/hmedown_20080317-20080402_170823/1207181304/1240249987-1587405187/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Guru Guides (3/08 - 1143)</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      403871C3-A324-21B8-1D8D-826834E758EB\r\n" + 
//					"     </Uuid> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/kidzone_v1.0-20100622-20100622_141902/1277241543/1289849391-1605209391/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/kidzone_v1.0-20100622-20100622_141902/1277241543/1289849391-1605209391/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>KidZone (3/8-1143)</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      5D524EE8-B977-1DC8-CD30-39BF60E7749C\r\n" + 
//					"     </Uuid> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/kidzone_v1.0-20100622-20100622_141902/1277241543/1289849349-1605209349/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/kidzone_v1.0-20100622-20100622_141902/1277241543/1289849349-1605209349/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Live365</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      9B26EF4C-BF08-A9A5-F18E-F898DA9ADEB6\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      35\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/live365-1.0_2010.08.02-2340-20100802_234348/1280817829/1343239177-1658599177/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/live365-1.0_2010.08.02-2340-20100802_234348/1280817829/1343239177-1658599177/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Music Videos from Music Choice</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      7FDF1BEB-AF92-B8B6-6269-FAE44C8F887D\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      49\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/musicchoice-1.0-2010.08.04-0632-20100804_161353/1280963634/1328046275-1643406275/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/musicchoice-1.0-2010.08.04-0632-20100804_161353/1280963634/1328046275-1643406275/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Podcaster</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      6838E9A3-F968-683E-110E-6591066452A1\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      45\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/podcast-20100730-qbert-20100730_180250/1280538174/1343239384-1658599384/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/podcast-20100730-qbert-20100730_180250/1280538174/1343239384-1658599384/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>SameGame</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      DC47F5CD-C97F-1C8F-EA91-CDDB4CFA1EFA\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      65\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/samegame_20071012_01-20080402_173700/1207183021/1343239500-1658599500/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/samegame_20071012_01-20080402_173700/1207183021/1343239500-1658599500/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Skull &amp; Bones</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      1A73F688-60B2-561D-BBEA-354BB36D3DA4\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      80\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/skullbones_20071012-20080402_173743/1207183066/1343239554-1658599554/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/skullbones_20071012-20080402_173743/1207183066/1343239554-1658599554/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Survey App</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      5A695917-238E-4218-8066-D4713A314063\r\n" + 
//					"     </Uuid> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/survey_20080530-01_signed-20080530_150815/1212185297/1289848835-1605208835/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/survey_20080530-01_signed-20080530_150815/1212185297/1289848835-1605208835/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Swivel 2.8 U.S. Jump Target  (03/16/11-0631)</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      49C791F9-7D48-38AB-8177-9C756D4EE2C0\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      50\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/swivel-2.8.3-2012-05-08-0603at553106-20120508_094009/1336495211/1336595307-1651955307/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/swivel-2.8.3-2012-05-08-0603at553106-20120508_094009/1336495211/1336595307-1651955307/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Swivel 2.8 (03/16/11-0631)</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      2BD40EF7-D9D2-3CD4-0130-6C18855CE898\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      50\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/swivel-2.8.3-2012-05-08-0603at553106-20120508_094009/1336495211/1336595232-1651955232/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/swivel-2.8.3-2012-05-08-0603at553106-20120508_094009/1336495211/1336595232-1651955232/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>TiVoCast 2.8 (2011.03.16-0115)</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      19452283-D00B-C411-595D-447ABF62F37C\r\n" + 
//					"     </Uuid> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/tivocast-2.8-2012-06-26-0603at563743-20120626_120132/1340737293/1344881428-1660241428/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/tivocast-2.8-2012-06-26-0603at563743-20120626_120132/1340737293/1344881428-1660241428/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <Hidden /> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Video Player</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      863CB78F-EFDD-4106-B572-51733983DC76\r\n" + 
//					"     </Uuid> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       https://204.176.49.65:7287/videoplayer-0023-release-sst-20090716_122445/1247772287/1333658049-1649018049/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/videoplayer-0023-release-sst-20090716_122445/1247772287/1333658049-1649018049/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"   <Item> \r\n" + 
//					"    <Details> \r\n" + 
//					"     <ContentType>\r\n" + 
//					"      application/x-hme\r\n" + 
//					"     </ContentType> \r\n" + 
//					"     <SourceFormat>\r\n" + 
//					"      x-container/folder\r\n" + 
//					"     </SourceFormat> \r\n" + 
//					"     <Title>Wordsmith</Title> \r\n" + 
//					"     <Uuid>\r\n" + 
//					"      41731A91-6B5D-7E7D-C495-9A1F399965C7\r\n" + 
//					"     </Uuid> \r\n" + 
//					"     <X-TiVo-TopLevelDisplayOrderHint>\r\n" + 
//					"      70\r\n" + 
//					"     </X-TiVo-TopLevelDisplayOrderHint> \r\n" + 
//					"    </Details> \r\n" + 
//					"    <Links> \r\n" + 
//					"     <Content> \r\n" + 
//					"      <ContentType>\r\n" + 
//					"       application/x-hme\r\n" + 
//					"      </ContentType> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/wordsmith_20071012-20080402_180020/1207184421/1343239627-1658599627/clientip_67.0.59.230/\r\n" + 
//					"      </Url> \r\n" + 
//					"     </Content> \r\n" + 
//					"     <CustomIcon> \r\n" + 
//					"      <Url>\r\n" + 
//					"       http://204.176.49.116:7288/wordsmith_20071012-20080402_180020/1207184421/1343239627-1658599627/icon.png\r\n" + 
//					"      </Url> \r\n" + 
//					"     </CustomIcon> \r\n" + 
//					"    </Links> \r\n" + 
//					"   </Item> \r\n" + 

//					"  </TiVoContainer> \r\n" + 
//					"";
