package com.blackledge.david.tivo;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Map;

import com.tivo.hme.bananas.BApplication;
import com.tivo.hme.bananas.BScreen;
import com.tivo.hme.bananas.BView;
import com.tivo.hme.bananas.IBananas;
import com.tivo.hme.interfaces.IContext;
import com.tivo.hme.sdk.HmeEvent;

public class PassThroughApp extends V49BApplication {

	public static String URI = "passthrough";
	
	// transmits v49 so we get keyboard key codes
	public static class PassThroughAppFactory extends V49BApplicationFactory {
	}
	
	public static class PassThroughScreen extends BScreen {

		private String arg;
		private BView text;
		private AppPassThroughView appView;
		private Map map;

		public PassThroughScreen(BApplication app, String arg, Map map) {
			super(app);
			this.arg = arg;
			this.map = map;
		}
		
		@Override
		public boolean handleEnter(Object arg, boolean isReturn) {
			V49BApplication app = (V49BApplication)getBApp();
			if(!isReturn) {
				this.text = new BView(this.getAbove(), app.getSafeActionHorizontal(), app.getSafeActionVertical(), getWidth()-2*app.getSafeActionHorizontal(), 50);
				this.text.setResource(createText("default-20.font",
						Color.yellow, "Press SELECT then passthrough appears here."));
				
				this.appView = new AppPassThroughView(this.getNormal(), 0, 0, getWidth(), getHeight());
				
				ArrayList<Integer> keys = new ArrayList<Integer>();
				keys.add(KEY_CLEAR);
				this.appView.setKeyPressOverrides(keys, this);
				
				setFocusDefault(appView);
			}
			super.handleEnter(arg, isReturn);
			return true;
		}

		@Override
		public boolean handleKeyPress(int code, long rawcode) {
			
			if(code == KEY_SELECT && !this.appView.isStarted()) {
				this.text.setResource(createText("default-20.font",
						Color.yellow, "Passthrough appears here."));
				//NOTE: both video apps only use: "transport keys" (play/pause/slow/ff/rw/adv/rep) clear, info, left, and right.
				// left quits, and right (info area changes) is a little different between them.
//				this.appView.setConsumingEvents(false); // we want to see events, too.
				this.appView.startApp(this.arg, this.map);
				return true;
			}
			if(code == KEY_CLEAR) {
				getBApp().setActive(false);
				return true;
			}
			return super.handleKeyPress(code, rawcode);
		}
		
		
		@Override
		public boolean handleEvent(HmeEvent event) {
			switch (event.getOpCode()) {
			case EVT_KEY: {
				if (this.appView.getResource() != null) {
					HmeEvent.Key ir = (HmeEvent.Key) event;
					if (!this.appView.isOverriddenKey(ir.getCode(), ir.getRawCode())) {
						if (this.text.getResource() != null)
							this.text.getResource().remove();
						this.text.setResource(createText("default-20.font",
								Color.yellow, ir.toString()));
						return true;
					}
				}
				break;
			}
			}
			return super.handleEvent(event);
		}

		public void handleActivation(boolean active) {
			
			if(!active) {
				System.out.println("**INACTIVE");
//				if(this.appView.getResource() != null)
//					this.appView.getResource().setActive(false);
			}
			else {
				System.out.println("**ACTIVE**");
			}
		}
	}

	
	
	
	private String arg;
	
	@Override
	protected void initBeforeResolutionSet(IContext context) {
		super.initBeforeResolutionSet(context);
		arg = context.getConnectionAttribute("arg");
	}
	
	@Override
	protected void showFirstPage() {
		push(new PassThroughScreen(this, arg, null), IBananas.TRANSITION_NONE);
	}

	@Override
	public boolean handleEvent(HmeEvent event) {
		switch (event.getOpCode()) {
        case EVT_APP_INFO: {
        	System.out.println("* * * APP: "+event);
            HmeEvent.ApplicationInfo info = (HmeEvent.ApplicationInfo)event;
            if (info.getMap().get("active") != null && getCurrentScreen() != null) {
                ((PassThroughScreen) getCurrentScreen()).handleActivation("true".equals(info.getMap().get("active")));
            }
            break;
        }
//		case EVT_KEY: {
//			if (this.appView.getResource() != null) {
//				HmeEvent.Key ir = (HmeEvent.Key) event;
//				if (this.text.getResource() != null)
//					this.text.getResource().remove();
//				this.text.setResource(createText("default-20.font",
//						Color.yellow, ir.toString()));
//				this.appView.getResource().sendEvent(event);
//			}
//			break;
//		}
		}
		return super.handleEvent(event);
	}
}
