package com.blackledge.david.tivo;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.HashMap;

import com.tivo.hme.sdk.IHmeProtocol;

//Roamio Series DVRs
//
//
//First 3 digits of the TiVo Service Number (TSN)	Model Number(s)	Manufacturer and model	Current software version
//840	TCD840200	Roamio Pro	20.3.7.1a
//848	TCD848000	Roamio Plus	20.3.7.1a
//846	TCD846500	4-Tuner Roamio	20.3.7.1a
//TiVo Mini
//
//
//First 3 digits of the TiVo Service Number (TSN)	Manufacturer and model	Current software version
//A92	TiVo Mini	20.3.7.1a
//TiVo Stream
//
//
//First 3 digits of the TiVo Service Number (TSN)	Manufacturer and model	Current software version
//A94	TiVo Stream	19.1.2a
//Premiere Series DVRs
//
//
//First 3 digits of the TiVo Service Number (TSN)	Model Number(s)	Manufacturer and model	Current software version
//746	TCD74650	Premiere (Retail - 500GB)	20.3.1
//750	TCD75050	Premiere 4 (Retail)	20.3.1
//758	TCD75825	Premiere XL4/Elite (Retail)	20.3.1
//746	TCD74632	Premiere (Retail - 320GB)	20.3.1
//748	TCD74800	Premiere XL (Retail)	20.3.1
//Series3 DVRs
//
//
//First 3 digits of the TiVo Service Number (TSN)	Model Number(s)	Manufacturer and model	Current software version
//658	TCD658000	TiVo HD XL	11.0k/11.0m
//652	TCD652160	TiVo HD	11.0k/11.0m
//648	TCD648250b	TiVo Series3 HD	11.0k/11.0m
//Series1 & Series2 DVRs
//
//
//First 3 digits of the TiVo Service Number (TSN)	Model Number(s)	Manufacturer and model	Current software version
//649	TCD649080
//TCD649180	TiVo Series2 Dual Tuner	9.3.2/9.3.2b/9.3.2c
//595	DRT400
//DRT800
//DRT2000	Humax Series2 DVD Writer	9.3.2
//590	T800
//T2500	Humax Series2	9.3.2
//565	RS-TX20
//RS-TX60	Toshiba Series2 DVD Writer	9.3.2
//540
//542	TCD540040
//TCD540080
//TCD540040
//TCD540080
//TCD540140	TiVo Series2	9.3.2/9.3.2b
//542	TCD542040
//TCD542080	TiVo Series2	9.3.2/9.3.2b
//275	box-810H-S
//box-57H	Pioneer Series2 DVD Recorder	9.3.2
//First 3 digits of the TiVo Service Number (TSN)	Model Number(s)	Manufacturer and model	Current software version
//264	SD-H400	Toshiba Series2 DVD Player	9.3.2
//240	TCD240040
//TCD240080
//TCD240140	TiVo Series2	9.3.2/9.3.2b
//230	TCD230040	AT&T/TiVo Series2	9.3.2/9.3.2b
//140	TCD140060
//TCD140080	TiVo Series2	9.3.2/9.3.2b
//130	TCD130040	AT&T Series2	9.3.2/9.3.2b
//110	SVR3000	Sony Series2	9.3.2/9.3.2b
//010	SVR2000	Sony Series1	7.3
//002	HDR112
//HDR212
//HDR312
//HDR612	Phillips Series1	7.3
//000	HDR112
//HDR312	Phillips Series1	7.3
/**
 * TSN number information: http://support.tivo.com/app/answers/detail/a_id/311 <BR/>
 * other (especially DirectTV) information: http://www.tivopedia.com/tivo-service-number.php<BR/>
 * capacities and other information: http://en.wikipedia.org/wiki/TiVo_digital_video_recorders<BR/>
 * @author David
 *
 */
public class TiVoModelInfo implements Cloneable {
	public static enum SERIES {
		SERIES_1, SERIES_2, SERIES_3, SERIES_3_TIVO_HD, 
		SERIES_4_PREMIERE, SERIES_4_STREAM, SERIES_4_VM, SERIES_4_XG1, SERIES_4_ONO,
		SERIES_5_MINI, SERIES_5_ROAMIO
	};

	private String _tsn;
	private long _tsn_i;
	private String _tsn_prefix;
	private int _tsn_prefix_i;
	
	private String _name;
//	private String _swver;
	private SERIES _series;
	private int _tuners;
	private int _capacity;
	private boolean _hd;
	private boolean _dvd;
	private boolean _directv;
	private boolean _stream;
	private boolean _hme;
	private boolean _videoStream;
	private boolean _rpc;
	private boolean _mind;
	
	private TiVoRemoteInfo _defaultRemote;
	
	private final static String modelsFilename = "com/blackledge/david/tivo/TiVo Model Information - Sheet1.csv";
	private final static HashMap<Integer, TiVoModelInfo> models = new HashMap<Integer, TiVoModelInfo>();
	
	public static synchronized TiVoModelInfo getInfo(String tsn) {
//		System.out.println("getting for "+tsn);
		if(models.size() == 0) {
			InputStream fileStream = TiVoModelInfo.class.getClassLoader().getResourceAsStream(modelsFilename);
			if(fileStream == null) {
				errorOut(modelsFilename+" file failed");
				return null;
			}
			BufferedReader reader = new BufferedReader(new InputStreamReader(fileStream));
			try {
				for (String line = reader.readLine(); line != null; line = reader.readLine()) {
					if(line.trim().length() == 0) continue;
					if(line.startsWith("#")) continue;
					if(line.startsWith(",,")) continue;
					//#TSN	name	swver	series	hd	tuners	streams to devices	(min) capacity	remote	directv	dvd	3rd party hme	rpc (per kmttg)	mind (per kmttg)	video streaming (HME)
					String[] fields = line.split(",");
//					System.out.println(fields.length+" - "+line);
					if(fields.length == 1) continue;
					int i = 0;
					
					try {

						String tsn_field = fields[i];
						i++;
						if(tsn_field.startsWith("0x")) tsn_field = tsn_field.substring(2);
						String tsn_prefix = tsn_field.substring(0, 3);
						int tsn_prefix_i = Integer.parseInt(tsn_prefix, 16);
						TiVoModelInfo model = new TiVoModelInfo(tsn_prefix, tsn_prefix_i);

						String nextField = nextField(fields, i);
						if(nextField != null)
							model._name = nextField;
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
//						if(nextField != null)
//							model._swver = nextField;
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._series = SERIES.valueOf(nextField);
						} catch(Exception e) {
							errorOut(e);
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._hd = Boolean.parseBoolean(nextField);
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._tuners = Integer.parseInt(nextField);
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._stream = Boolean.parseBoolean(nextField);
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._capacity = Integer.parseInt(nextField);
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._defaultRemote = TiVoRemoteInfo.getRemote(nextField.trim());
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._directv = Boolean.parseBoolean(nextField);
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._dvd = Boolean.parseBoolean(nextField);
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._hme = Boolean.parseBoolean(nextField);
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._rpc = Boolean.parseBoolean(nextField);
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._mind = Boolean.parseBoolean(nextField);
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						nextField = nextField(fields, i);
						try {
							if(nextField != null)
								model._videoStream = Boolean.parseBoolean(nextField);
						} catch(Exception e) {
						}
						i += 1 + (commaCount(nextField));

						models.put(model._tsn_prefix_i, model);

//						System.out.println(model._name);
					} catch(Exception e) {
//						errorOut(e);
						continue;
					}
				}
			} catch (IOException e) {
				e.printStackTrace();
				return null;
			}
		}
		
		try {
			String tsn_prefix = tsn.substring(0,3);
			int tsn_prefix_i = Integer.parseInt(tsn_prefix, 16);
			TiVoModelInfo model = models.get(tsn_prefix_i);
			if (model != null)
				return model.instance(tsn);
			else
				return null;
		} catch (Exception e) {
			return null;
		}
	}
	
	private static void errorOut(Object o) {
		System.err.println(o);
	}
	
	@Override
	public String toString() {
		return _tsn+":"+_series+":"+_name;
	}

	/** handle quote "escaping" of fields with commas*/
	private static String nextField(String[] fields, int i) {
		while (fields.length > i && fields[i].length() > 0) {
			String result = fields[i];
			if (result.startsWith("\"")) {
				while ((!result.endsWith("\"\"\"") && 
						(result.endsWith("\"\"") || !result.endsWith("\"")))
						&& fields.length > i + 1) {
					++i;
					result += "," + fields[i];
				}
				result = result.substring(1, result.length()-1);
				result = result.replace("\"\"", "\"");
			}
			return result;
		}
		return null;
	}

	private static int commaCount(String nextField) {
		if(nextField == null) return 0;
		int i = 0;
		int index = nextField.indexOf(',');
		while(index >= 0)
		{
			++i;
			index = nextField.indexOf(',',index+1);
		}
		return i;
	}

	public static void main(String[] args) {
		TiVoModelInfo sample = getInfo("748");
		System.out.println("------------");
		if(sample != null) {
			System.out.println(sample);
			TiVoRemoteInfo remote = sample.getDefaultRemote();
			if(remote != null) {
				System.out.println(remote.getKeyName(IHmeProtocol.KEY_ADVANCE));
				System.out.println(remote.getKeyIcon(IHmeProtocol.KEY_ADVANCE));
				System.out.println(remote.getKeyName(IHmeProtocol.KEY_ENTER));
				System.out.println(remote.getKeyIcon(IHmeProtocol.KEY_ENTER));
				System.out.println(remote.getKeyName(IHmeProtocol.KEY_OPT_ASPECT));
				System.out.println(remote.getKeyIcon(IHmeProtocol.KEY_OPT_ASPECT));
			}
		}
	}

	private TiVoModelInfo instance(String tsn) {
		TiVoModelInfo clone;
		try {
			clone = (TiVoModelInfo) this.clone();
		} catch (CloneNotSupportedException e) {
			errorOut(e);
			return null;
		}
		
		clone._tsn = tsn;
		try {
			clone._tsn_i = Long.parseLong(tsn, 16);
		} catch(Exception e) {
		}
		
		return clone;
	}
	
	private TiVoModelInfo(String tsn_prefix, int tsn_prefix_i) {
		_tsn_prefix = tsn_prefix;
		_tsn_prefix_i = tsn_prefix_i;
	}		
	
//	public TiVoModelInfo(String tsn) {
//		_tsn = tsn;
//		_tsn_i = Long.parseLong(tsn, 16);
//		_tsn_prefix = tsn.substring(0,3);
//		_tsn_prefix_i = Integer.parseInt(_tsn_prefix, 16);
//
//		// typical defaults
//		_tuners = 1;
//		_hd = false;
//		_dvd = false;
//		_directv = false;
//		_streams = false;
//		
//		_capacity = -1;
//		
//
//		switch(_tsn_prefix_i) {
//		case 0x840:	// Roamio Pro	20.3.7.1a
//			_series = SERIES.SERIES_5_ROAMIO;
//			_hd = true;
//			_tuners = 6;
//			_streams = true;
//			_capacity = 3072;
//			break;
//		case 0x848:	// Roamio Plus	20.3.7.1a
//			_series = SERIES.SERIES_5_ROAMIO;
//			_hd = true;
//			_tuners = 6;
//			_streams = true;//??
//			_capacity = 1024;
//			break;
//		case 0x846:	// 4-Tuner Roamio	20.3.7.1a
//			_series = SERIES.SERIES_5_ROAMIO;
//			_hd = true;
//			_tuners = 4;
//			_capacity = 500;
//			break;
//		
//		case 0xC00: // FIXME - details? - Virgin Media UK TiVos
//			break;
//			
//		case 0xA92:	// TiVo Mini
//			_series = SERIES.SERIES_5_MINI;//right?
//			_hd = true;
//			_tuners = 0;// no REAL tuners, but commandeers one
//			_capacity = 0;
//			break;
//		
//		case 0xA94:	// TiVo Stream
//			_series = SERIES.SERIES_4_STREAM;//right?
//			_hd = true;
//			_tuners = 0;
//			_streams = true;
//			_capacity = 0;
//			break;
//		
//		
//		case 0x750:	// Premiere 4 (Retail)	20.3.1
//			_series = SERIES.SERIES_4_PREMIERE;
//			_hd = true;
//			_tuners = 4;
//			_capacity = 500;
//			// premiere remote
//			_defaultRemote = TiVoRemoteInfo.PREMIERE_REMOTE;
//			break;
//		case 0x758:	// Premiere XL4/Elite (Retail)	20.3.1
//			_series = SERIES.SERIES_4_PREMIERE;
//			_hd = true;
//			_tuners = 4;
//			_capacity = 2048;
//			// premium glo remote
//			_defaultRemote = TiVoRemoteInfo.PREMIERE_GLO_REMOTE;
//			break;
//		case 0x746:	// Premiere (Retail - 320GB)	20.3.1
//			_series = SERIES.SERIES_4_PREMIERE;
//			_hd = true;
//			_tuners = 2;
//			_capacity = 320;
//			// premiere remote
//			_defaultRemote = TiVoRemoteInfo.PREMIERE_REMOTE;
//			break;
//		case 0x748:	// Premiere XL (Retail)	20.3.1
//			_series = SERIES.SERIES_4_PREMIERE;
//			_hd = true;
//			_tuners = 2;
//			_capacity = 1024;
//			// premium glo remote
//			_defaultRemote = TiVoRemoteInfo.PREMIERE_GLO_REMOTE;
//			break;
//		
//		case 0x663: // FIXME - DETAILS?- seen in Enter Webz logs - Australian and New Zealand models
//			_series = SERIES.SERIES_3_TIVO_HD;
//			_hd = true;
//			_tuners = 2; // right?
//			_capacity = 160; // right?
//			// hd remote
//			_defaultRemote = TiVoRemoteInfo.TIVO_HD_REMOTE; // right?
//			break;
//		case 0x653: // FIXME - details? - TiVo in Mexico City via CABLEVISION
//			_series = SERIES.SERIES_3_TIVO_HD;
//			_hd = true;
//			_tuners = 2; // right?
//			_capacity = 160; // right?
//			// hd remote
//			_defaultRemote = TiVoRemoteInfo.TIVO_HD_REMOTE; // right?
//			break;
//			
//		case 0x658:	// TiVo HD XL	11.0k/11.0m
//			_series = SERIES.SERIES_3_TIVO_HD;
//			_hd = true;
//			_tuners = 2;
//			_capacity = 1024;
//			// glo remote
//			_defaultRemote = TiVoRemoteInfo.SERIES3_GLO_REMOTE;
//			break;
//		case 0x652:	// TiVo HD	11.0k/11.0m
//			_series = SERIES.SERIES_3_TIVO_HD;
//			_hd = true;
//			_tuners = 2;
//			_capacity = 160;
//			// hd remote
//			_defaultRemote = TiVoRemoteInfo.TIVO_HD_REMOTE;
//			break;
//		case 0x648:	// TiVo Series3 HD	11.0k/11.0m
//			_series = SERIES.SERIES_3;
//			_hd = true;
//			_tuners = 2;
//			_capacity = 250;
//			//glo remote
//			_defaultRemote = TiVoRemoteInfo.SERIES3_GLO_REMOTE;
//			break;
//		
//		case 0x627: // DirecTV (THR22)
//			_hd = true;
//			_directv = true;
//			//FIXME fill this in
//			break;
//		
//		case 0x649:	// TiVo Series2 Dual Tuner	9.3.2/9.3.2b/9.3.2c
//			_series = SERIES.SERIES_2;
//			_tuners = 2;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//		case 0x595:	// Humax Series2 DVD Writer	9.3.2
//			_series = SERIES.SERIES_2;
//			_dvd = true;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//		case 0x590:	// Humax Series2	9.3.2
//			_series = SERIES.SERIES_2;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//		case 0x565:	// Toshiba Series2 DVD Writer	9.3.2
//			_series = SERIES.SERIES_2;
//			_dvd = true;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//		case 0x540:
//			_series = SERIES.SERIES_2;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//		case 0x542:	// TiVo Series2	9.3.2/9.3.2b
//			_series = SERIES.SERIES_2;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//			
//		case 0x521: // DIRECTV (R10)
//			//FIXME fill this in
//			_directv = true;
//			break;
//		case 0x381: // Samsung (SIR-S...)
//			//FIXME fill this in
//			_directv = true;
//			break;
//		case 0x357: // Hughes / DIRECTV (HR10-250)
//			//FIXME fill this in
//			_hd = true;
//			_directv = true;
//			break;
//		case 0x351: // Hughes (SD-DVR...)
//			//FIXME fill this in
//			_directv = true;
//			break;
//		case 0x321: // RCA (DVR...)
//			//FIXME fill this in
//			_directv = true;
//			break;
//		case 0x301: // Philips (DSR...)
//			//FIXME fill this in
//			_directv = true;
//			break;
//			
//		case 0x275:	// Pioneer Series2 DVD Recorder	9.3.2
//			_series = SERIES.SERIES_2;
//			_dvd = true;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//		
//		case 0x264:	// Toshiba Series2 DVD Player	9.3.2
//			_series = SERIES.SERIES_2;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//		case 0x240:	// TiVo Series2	9.3.2/9.3.2b
//			_series = SERIES.SERIES_2;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//		case 0x230:	// AT&T/TiVo Series2	9.3.2/9.3.2b
//			_series = SERIES.SERIES_2;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//			
//		case 0x151: // Hughes (HDVR2)
//			//FIXME fill this in
//			_directv = true;
//			break;
//			
//			
//		case 0x140:	// TiVo Series2	9.3.2/9.3.2b
//			_series = SERIES.SERIES_2;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//		case 0x130:	// AT&T Series2	9.3.2/9.3.2b
//			_series = SERIES.SERIES_2;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//			
//		case 0x121: // RCA (DVR39)
//			//FIXME fill this in
//			_directv = true;
//			break;
//			
//			
//		case 0x110:	// Sony Series2	9.3.2/9.3.2b
//			_series = SERIES.SERIES_2;
//			_capacity = 40;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES2_REMOTE;//right?
//			break;
//
//		case 0x101: // Philips DSR...
//			//FIXME fill this in
//			_directv = true;
//			break;
//		case 0x031: // Hughes (GXCEBOT(D))
//			//FIXME fill this in
//			_directv = true;
//			break;
//		case 0x011: // Sony (SAT T-60)
//			//FIXME fill this in
//			_directv = true;
//			break;
//			
//		case 0x023: // FIXME - details? - BSkyB Series1 UK TiVos ( deactivated)
//			_series = SERIES.SERIES_1;
//			break;
//			
//			
//		case 0x010:	// Sony Series1	7.3
//			_series = SERIES.SERIES_1;
//			_capacity = 30;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES1_REMOTE;
//			break;
//		case 0x002:	// Phillips Series1	7.3
//			_series = SERIES.SERIES_1;
//			_capacity = 30;//right?
//			_defaultRemote = TiVoRemoteInfo.SERIES1_REMOTE;
//			break;
//			
//		case 0x001: // Philips (DSR...)
//			//FIXME fill this in
//			_directv = true;
//			_capacity = 40;//right?
//			break;
//			
//			
//		case 0x000:	// Philips Series1	7.3
//			_series = SERIES.SERIES_1;
//			_capacity = 14;
//			_defaultRemote = TiVoRemoteInfo.SERIES1_REMOTE;
//			break;
//
//		}
//	}

	public String getTSN() {
		return _tsn;
	}
	
	public String getTSNPrefix() {
		return _tsn_prefix;
	}
	
	public SERIES getSeries() {
		return _series;
	}
	
	public int getSeriesNum() {
		switch(_series) {
		case SERIES_1:
			return 1;
		case SERIES_2:
			return 2;
		case SERIES_3:
		case SERIES_3_TIVO_HD:
			return 3;
		case SERIES_4_PREMIERE:
		case SERIES_4_STREAM:
		case SERIES_4_ONO:
		case SERIES_4_VM:
		case SERIES_4_XG1:
			return 4;
		case SERIES_5_MINI:
		case SERIES_5_ROAMIO:
			return 5;
		}
		// not really possible unless I forget to add a new case.
		return -1;
	}

	public String getSeriesName() {
		return getSeriesName(_series);
	}
	
	/**
	 * The common name for the series. <BR>
	 * (returns "TiVo HD" for all Series 3 models, "Virgin Media" for VM's
	 * Premieres, and "Premiere" for ONO(Spain) and XG1(Scandinavia) models). <BR>
	 * Returns null for a null/unknown series.
	 */
	public static String getSeriesName(SERIES series) {
		switch(series) {
		case SERIES_1: return "Series 1";
		case SERIES_2: return "Series 2";
		case SERIES_3: //return "Series 3";
		case SERIES_3_TIVO_HD: return "TiVo HD";
		case SERIES_4_VM: return "Virgin Media";
		case SERIES_4_ONO: //return "ONO";
		case SERIES_4_XG1: //return "XG1";
		case SERIES_4_PREMIERE: return "Premiere";
		case SERIES_4_STREAM: return "TiVo Stream";
		case SERIES_5_MINI: return "TiVo Mini";
		case SERIES_5_ROAMIO: return "Roamio";
		}
		return null;
	}
	
	
	/** the Descriptive name of this model */
	public String getName() {
		return _name;
	}

	public boolean isDirectv() {
		return _directv;
	}

	/**
	 * Has a DVD player or recorder
	 */
	public boolean isDvd() {
		return _dvd;
	}

	/**
	 * Capable of processing HD video
	 */
	public boolean isHd() {
		return _hd;
	}
	
	/**
	 * Capable of running 3rd-party HME apps
	 */
	public boolean isHme() {
		return _hme;
	}
	
	public boolean isMind() {
		return _mind;
	}
	
	public boolean isRpc() {
		return _rpc;
	}
	
	/**
	 * Capable of streaming video to a non-TiVo device
	 */
	public boolean isStream() {
		return _stream;
	}
	
	/**
	 * Capable of streaming video (i.e. via HME)
	 */
	public boolean isVideoStream() {
		return _videoStream;
	}
	
	/**
	 * Minimum default capacity for this model. 0 if this device cannot store
	 * video itself.
	 */
	public int getCapacity() {
		return _capacity;
	}

	/**
	 * Number of TV tuners for this model. 0 for e.g. TiVo Stream or
	 * Tuner-allocating devices e.g. TiVo Mini.
	 */
	public int getTuners() {
		return _tuners;
	}
	
	/**
	 * Whether this model can do streaming. True for TiVo Stream and models with
	 * built-in streaming such as some Roamio models.
	 */
	public boolean isStreaming() {
		return _stream;
	}

	/**
	 * Whether or not this model supports HD.
	 */
	public boolean isHD() {
		return _hd;
	}

	/**
	 * Whether or not this model includes a DVD player or recorder.
	 */
	public boolean isDVD() {
		return _dvd;
	}
	
	public TiVoRemoteInfo getDefaultRemote() {
		return _defaultRemote;
	}

	/**
	 * Shortcut to getDefaultRemote().getKeyName(). If getDefaultRemote() is
	 * null, uses the {@link TiVoRemoteInfo#PREMIERE_REMOTE}.
	 * See also {@link #getRemoteKeyIcon(int)}
	 * @param keyCode
	 * @return button text of that key on the remote, or null if that
	 *         button doesn't exist on the remote or only has an icon.
	 */
	public String getRemoteKeyName(int keyCode) {
		TiVoRemoteInfo remote = getDefaultRemote();
		if (remote == null)
			remote = TiVoRemoteInfo.getDefaultRemote();
		if(remote == null) {
			return null;
		}
		return remote.getKeyName(keyCode);
	}
	
	public String getRemoteKeyIcon(int keyCode) {
		TiVoRemoteInfo remote = getDefaultRemote();
		if (remote == null)
			remote = TiVoRemoteInfo.getDefaultRemote();
		if(remote == null) {
			return null;
		}
		return remote.getKeyIcon(keyCode);
	}

}
